import os
from pathlib import Path
from types import SimpleNamespace

import logging
import moderngl as mgl
import numpy as np

import pytest

os.environ.setdefault("QT_QPA_PLATFORM", "offscreen")

pytest.importorskip("PySide6")

from PySide6 import QtCore, QtGui, QtWidgets

from topovec.mgl.camera import Camera
from topovec.core import BooleanProperty, ListProperty, System, UpdateMask
from topovec.viewer.document import StateEntry, ViewerDocument
from topovec.viewer.qt.main_window import TvViewWindow
from topovec.viewer.qt.theme import THEME_SETTING_KEY
from topovec.viewer.qt.viewport import QtOpenGLLoader, QtViewport, QtViewportWindow
from topovec.viewer.session import ViewerSession, camera_state_signature
from topovec.viewer.qt.widgets import (
    LogPanel,
    NUMERIC_INPUT_DECIMALS,
    PROPERTY_CONTROL_HEIGHT,
    PROPERTY_RESET_WIDTH,
    PropertyPanel,
)

REAL_QSETTINGS = QtCore.QSettings


@pytest.fixture(scope="module")
def qapp():
    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication([])
    return app


@pytest.fixture(autouse=True)
def topovec_view_settings_path(monkeypatch, tmp_path):
    settings_path = tmp_path / "topovec.view.ini"

    def _make_settings(organization, application):
        assert organization == "topovec"
        assert application == "topovec.view"
        return REAL_QSETTINGS(str(settings_path), REAL_QSETTINGS.Format.IniFormat)

    monkeypatch.setattr("topovec.viewer.qt.main_window.QtCore.QSettings", _make_settings)
    return settings_path


class _QtSceneBase:
    def __init__(self, mglctx=None, system=None):
        self._mglctx = mglctx
        self._system = system
        self.camera = Camera()
        self.flag = False
        self.uploaded_state = None
        self.release_count = 0

    @property
    def system(self):
        return self._system

    @classmethod
    def supports_system(cls, system):
        return True

    def upload(self, state=None):
        self.uploaded_state = state

    def list_properties(self):
        return {
            "Render": [
                BooleanProperty(self, "flag", title="Flag"),
            ]
        }

    def render_to(self, target):
        return None

    def release(self):
        self.release_count += 1


class _QtLayerScene(_QtSceneBase):
    pass


class _QtBallScene(_QtSceneBase):
    pass


class _QtScopedVectorScene(_QtSceneBase):
    SHAPES = ["Cone", "Double cone", "Cylinder", "Arrow"]

    def __init__(self, *args, mutation_state=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._mutation_state = mutation_state if mutation_state is not None else {}
        self._mesh_shape = "Cone"

    def get_mesh_shape(self):
        return self._mesh_shape

    def set_mesh_shape(self, value):
        if not self._mutation_state.get("active", False):
            raise RuntimeError("qt scene mutation scope is not active")
        if value not in self.SHAPES:
            raise ValueError(f"Unsupported mesh shape: {value!r}")
        self._mesh_shape = str(value)
        return UpdateMask.REDRAW

    mesh_shape = property(get_mesh_shape, set_mesh_shape)

    def list_properties(self):
        return {
            "Vector": [
                ListProperty(
                    self,
                    "mesh_shape",
                    title="Shape",
                    values=list(self.SHAPES),
                )
            ]
        }


class _QtSchemaScene(_QtSceneBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mode = "alpha"

    def list_properties(self):
        return {
            "Render": [
                ListProperty(
                    self,
                    "mode",
                    title="Mode",
                    values=["alpha", "beta"],
                    reset=True,
                )
            ]
        }


def _install_qt_scene_registry(monkeypatch):
    registry = {
        "Layer": _QtLayerScene,
        "Ball": _QtBallScene,
    }
    monkeypatch.setattr("topovec.viewer.qt.main_window.SCENES", dict(registry))
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: dict(registry),
    )
    return registry


def _make_document(
    name: str,
    system,
    states: list[np.ndarray],
    *,
    format_name: str = "SAD",
) -> ViewerDocument:
    entries = [
        StateEntry(
            label=f"PATH[{index}]",
            index=(int(index),),
            load_state=(lambda s=np.array(state, copy=True): np.array(s, copy=True)),
            metadata={"index": int(index)},
        )
        for index, state in enumerate(states)
    ]
    return ViewerDocument(
        path=Path(f"/tmp/{name}"),
        label=name,
        format_name=format_name,
        system=system,
        entries=entries,
        metadata={},
    )


def _read_recent_settings_paths(settings_path: Path) -> list[str]:
    settings = REAL_QSETTINGS(str(settings_path), REAL_QSETTINGS.Format.IniFormat)
    raw = settings.value("recent_files", [])
    if raw is None:
        return []
    if isinstance(raw, str):
        return [raw]
    return [str(value) for value in raw]


def _read_theme_setting(settings_path: Path) -> str | None:
    settings = REAL_QSETTINGS(str(settings_path), REAL_QSETTINGS.Format.IniFormat)
    value = settings.value(THEME_SETTING_KEY)
    return None if value is None else str(value)


def _menu_for_title(window: TvViewWindow, title: str) -> QtWidgets.QMenu:
    for menu in window.menuBar().findChildren(QtWidgets.QMenu):
        if menu.title() == title:
            return menu
    raise AssertionError(f"Menu {title!r} not found")


def test_property_panel_builds_generated_controls(qapp):
    panel = PropertyPanel()
    panel.set_snapshot(
        {
            "property_groups": [
                {
                    "id": "Render",
                    "title": "Render",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "numeric",
                            "title": "Level",
                            "field": "level",
                            "value": 0.5,
                            "min": 0.0,
                            "max": 1.0,
                            "count": 5,
                            "step": 0.25,
                            "input_kind": "number",
                            "row_kind": "inline",
                            "default_value": 0.5,
                            "is_default": True,
                            "can_reset": True,
                        },
                        {
                            "index": 1,
                            "kind": "boolean",
                            "title": "Flag",
                            "field": "flag",
                            "value": True,
                            "input_kind": "switch",
                            "row_kind": "inline",
                            "default_value": True,
                            "is_default": True,
                            "can_reset": True,
                        },
                        {
                            "index": 2,
                            "kind": "list",
                            "title": "Mode",
                            "field": "mode",
                            "value": "alpha",
                            "values": ["alpha", "beta"],
                            "selected_index": 0,
                            "input_kind": "dropdown",
                            "row_kind": "inline",
                            "default_value": "alpha",
                            "is_default": True,
                            "can_reset": True,
                        },
                    ],
                }
            ]
        }
    )

    assert not panel.findChildren(QtWidgets.QSlider)
    assert panel.findChildren(QtWidgets.QDoubleSpinBox)
    assert panel.findChildren(QtWidgets.QCheckBox)
    assert panel.findChildren(QtWidgets.QComboBox)


def test_property_panel_sections_start_collapsed(qapp):
    panel = PropertyPanel()
    panel.set_snapshot(
        {
            "property_groups": [
                {
                    "id": "Render",
                    "title": "Render",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "boolean",
                            "title": "Flag",
                            "field": "flag",
                            "value": False,
                            "default_value": False,
                            "is_default": True,
                            "can_reset": True,
                        }
                    ],
                }
            ]
        }
    )

    section = panel.widget().layout().itemAt(0).widget()
    toggle = section.layout().itemAt(0).widget()
    content = section.layout().itemAt(1).widget()

    assert toggle.isChecked() is False
    assert content.isHidden() is True


def test_property_panel_expanding_one_section_collapses_others(qapp):
    panel = PropertyPanel()
    panel.set_snapshot(
        {
            "property_groups": [
                {
                    "id": "Render",
                    "title": "Render",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "boolean",
                            "title": "Flag",
                            "field": "flag",
                            "value": False,
                            "default_value": False,
                            "is_default": True,
                            "can_reset": True,
                        }
                    ],
                },
                {
                    "id": "Vector",
                    "title": "Vector",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "boolean",
                            "title": "Visible",
                            "field": "visible",
                            "value": True,
                            "default_value": True,
                            "is_default": True,
                            "can_reset": True,
                        }
                    ],
                },
            ]
        }
    )

    first = panel.widget().layout().itemAt(0).widget()
    second = panel.widget().layout().itemAt(1).widget()

    first_toggle = first.layout().itemAt(0).widget()
    second_toggle = second.layout().itemAt(0).widget()
    first_content = first.layout().itemAt(1).widget()
    second_content = second.layout().itemAt(1).widget()

    first_toggle.setChecked(True)
    assert first_toggle.isChecked() is True
    assert first_content.isHidden() is False
    assert second_toggle.isChecked() is False
    assert second_content.isHidden() is True

    second_toggle.setChecked(True)
    assert second_toggle.isChecked() is True
    assert second_content.isHidden() is False
    assert first_toggle.isChecked() is False
    assert first_content.isHidden() is True


def test_property_panel_updates_in_place_without_emitting_changes(qapp):
    panel = PropertyPanel()
    emitted = []
    snapshot = {
        "property_groups": [
            {
                "id": "Render",
                "title": "Render",
                "properties": [
                    {
                        "index": 0,
                        "kind": "numeric",
                        "title": "Level",
                        "field": "level",
                        "value": 0.5,
                        "min": 0.0,
                        "max": 1.0,
                        "count": 5,
                        "step": 0.25,
                        "input_kind": "number",
                        "row_kind": "inline",
                        "default_value": 0.5,
                        "is_default": True,
                        "can_reset": True,
                    },
                    {
                        "index": 1,
                        "kind": "boolean",
                        "title": "Flag",
                        "field": "flag",
                        "value": True,
                        "input_kind": "switch",
                        "row_kind": "inline",
                        "default_value": True,
                        "is_default": True,
                        "can_reset": True,
                    },
                    {
                        "index": 2,
                        "kind": "list",
                        "title": "Mode",
                        "field": "mode",
                        "value": "alpha",
                        "values": ["alpha", "beta"],
                        "selected_index": 0,
                        "input_kind": "dropdown",
                        "row_kind": "inline",
                        "default_value": "alpha",
                        "is_default": True,
                        "can_reset": True,
                    },
                ],
            }
        ]
    }
    updated = {
        "property_groups": [
            {
                "id": "Render",
                "title": "Render",
                "properties": [
                    {
                        "index": 0,
                        "kind": "numeric",
                        "title": "Level",
                        "field": "level",
                        "value": 0.75,
                        "min": 0.0,
                        "max": 1.0,
                        "count": 5,
                        "step": 0.25,
                        "input_kind": "number",
                        "row_kind": "inline",
                        "default_value": 0.5,
                        "is_default": False,
                        "can_reset": True,
                    },
                    {
                        "index": 1,
                        "kind": "boolean",
                        "title": "Flag",
                        "field": "flag",
                        "value": False,
                        "input_kind": "switch",
                        "row_kind": "inline",
                        "default_value": True,
                        "is_default": False,
                        "can_reset": True,
                    },
                    {
                        "index": 2,
                        "kind": "list",
                        "title": "Mode",
                        "field": "mode",
                        "value": "beta",
                        "values": ["alpha", "beta"],
                        "selected_index": 1,
                        "input_kind": "dropdown",
                        "row_kind": "inline",
                        "default_value": "alpha",
                        "is_default": False,
                        "can_reset": True,
                    },
                ],
            }
        ]
    }
    panel.property_edited.connect(lambda *args: emitted.append(args))
    panel.set_snapshot(snapshot)

    spin_before = panel.findChild(QtWidgets.QDoubleSpinBox)
    checkbox_before = panel.findChild(QtWidgets.QCheckBox)
    combo_before = panel.findChild(QtWidgets.QComboBox)

    panel.set_snapshot(updated, rebuild=False)

    assert panel.findChild(QtWidgets.QDoubleSpinBox) is spin_before
    assert panel.findChild(QtWidgets.QCheckBox) is checkbox_before
    assert panel.findChild(QtWidgets.QComboBox) is combo_before
    assert spin_before.value() == 0.75
    assert spin_before.decimals() == NUMERIC_INPUT_DECIMALS
    assert spin_before.singleStep() == 0.25
    assert checkbox_before.isChecked() is False
    assert combo_before.currentText() == "beta"
    assert emitted == []


def test_property_panel_reset_button_emits_default_value(qapp):
    panel = PropertyPanel()
    emitted = []
    panel.property_edited.connect(lambda *args: emitted.append(args))
    panel.set_snapshot(
        {
            "property_groups": [
                {
                    "id": "Render",
                    "title": "Render",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "numeric",
                            "title": "Level",
                            "field": "level",
                            "value": 0.75,
                            "min": 0.0,
                            "max": 1.0,
                            "count": 5,
                            "step": 0.25,
                            "input_kind": "number",
                            "row_kind": "inline",
                            "default_value": 0.5,
                            "is_default": False,
                            "can_reset": True,
                        }
                    ],
                }
            ]
        }
    )

    reset_buttons = [
        button
        for button in panel.findChildren(QtWidgets.QToolButton)
        if button.property("inspectorRole") == "reset"
    ]

    assert len(reset_buttons) == 1
    assert reset_buttons[0].isEnabled() is True
    assert reset_buttons[0].text() == ""
    assert reset_buttons[0].icon().isNull() is False

    reset_buttons[0].click()

    assert emitted == [("Render", 0, 0.5, "level", "Level")]


def test_property_panel_numeric_step_commits_immediately(qapp):
    panel = PropertyPanel()
    emitted = []
    panel.property_edited.connect(lambda *args: emitted.append(args))
    panel.set_snapshot(
        {
            "property_groups": [
                {
                    "id": "Render",
                    "title": "Render",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "numeric",
                            "title": "Level",
                            "field": "level",
                            "value": 0.5,
                            "min": 0.0,
                            "max": 1.0,
                            "count": 5,
                            "step": 0.25,
                            "input_kind": "number",
                            "row_kind": "inline",
                            "default_value": 0.5,
                            "is_default": True,
                            "can_reset": True,
                        }
                    ],
                }
            ]
        }
    )

    spin = panel.findChild(QtWidgets.QDoubleSpinBox)

    spin.stepUp()
    spin.editingFinished.emit()

    assert emitted == [("Render", 0, 0.75, "level", "Level")]


def test_property_panel_numeric_text_commits_on_editing_finished_only(qapp):
    panel = PropertyPanel()
    emitted = []
    panel.property_edited.connect(lambda *args: emitted.append(args))
    panel.set_snapshot(
        {
            "property_groups": [
                {
                    "id": "Render",
                    "title": "Render",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "numeric",
                            "title": "Level",
                            "field": "level",
                            "value": 0.5,
                            "min": 0.0,
                            "max": 1.0,
                            "count": 5,
                            "step": 0.25,
                            "input_kind": "number",
                            "row_kind": "inline",
                            "default_value": 0.5,
                            "is_default": True,
                            "can_reset": True,
                        }
                    ],
                }
            ]
        }
    )

    spin = panel.findChild(QtWidgets.QDoubleSpinBox)
    spin.lineEdit().setText("0.75")
    spin.interpretText()

    assert emitted == []

    spin.editingFinished.emit()

    assert emitted == [("Render", 0, 0.75, "level", "Level")]


def test_property_panel_uses_reset_button_state_as_changed_signal(qapp):
    panel = PropertyPanel()
    panel.set_snapshot(
        {
            "property_groups": [
                {
                    "id": "Render",
                    "title": "Render",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "numeric",
                            "title": "Level",
                            "field": "level",
                            "value": 0.5,
                            "min": 0.0,
                            "max": 1.0,
                            "count": 5,
                            "step": 0.25,
                            "input_kind": "number",
                            "row_kind": "inline",
                            "default_value": 0.5,
                            "is_default": True,
                            "can_reset": True,
                        },
                        {
                            "index": 1,
                            "kind": "boolean",
                            "title": "Flag",
                            "field": "flag",
                            "value": False,
                            "input_kind": "switch",
                            "row_kind": "inline",
                            "default_value": True,
                            "is_default": False,
                            "can_reset": True,
                        },
                    ],
                }
            ]
        }
    )

    changed_labels = [
        label
        for label in panel.findChildren(QtWidgets.QLabel)
        if label.property("inspectorRole") == "changed"
    ]
    reset_buttons = [
        button
        for button in panel.findChildren(QtWidgets.QToolButton)
        if button.property("inspectorRole") == "reset"
    ]

    assert changed_labels == []
    assert len(reset_buttons) == 2
    assert reset_buttons[0].isEnabled() is False
    assert reset_buttons[1].isEnabled() is True


def test_property_panel_uses_uniform_control_heights(qapp):
    panel = PropertyPanel()
    panel.set_snapshot(
        {
            "property_groups": [
                {
                    "id": "Render",
                    "title": "Render",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "numeric",
                            "title": "Level",
                            "field": "level",
                            "value": 0.5,
                            "min": 0.0,
                            "max": 1.0,
                            "count": 5,
                            "step": 0.25,
                            "input_kind": "number",
                            "row_kind": "inline",
                            "default_value": 0.5,
                            "is_default": True,
                            "can_reset": True,
                        },
                        {
                            "index": 1,
                            "kind": "boolean",
                            "title": "Flag",
                            "field": "flag",
                            "value": True,
                            "input_kind": "switch",
                            "row_kind": "inline",
                            "default_value": True,
                            "is_default": True,
                            "can_reset": True,
                        },
                        {
                            "index": 2,
                            "kind": "list",
                            "title": "Mode",
                            "field": "mode",
                            "value": "alpha",
                            "values": ["alpha", "beta"],
                            "selected_index": 0,
                            "input_kind": "dropdown",
                            "row_kind": "inline",
                            "default_value": "alpha",
                            "is_default": True,
                            "can_reset": True,
                        },
                    ],
                }
            ]
        }
    )

    spin = panel.findChild(QtWidgets.QDoubleSpinBox)
    checkbox = panel.findChild(QtWidgets.QCheckBox)
    combo = panel.findChild(QtWidgets.QComboBox)
    reset = next(
        button
        for button in panel.findChildren(QtWidgets.QToolButton)
        if button.property("inspectorRole") == "reset"
    )

    assert spin.minimumHeight() == PROPERTY_CONTROL_HEIGHT
    assert checkbox.minimumHeight() == PROPERTY_CONTROL_HEIGHT
    assert combo.minimumHeight() == PROPERTY_CONTROL_HEIGHT
    assert reset.minimumHeight() == PROPERTY_CONTROL_HEIGHT
    assert reset.width() == PROPERTY_RESET_WIDTH


def test_topovec_view_window_constructs_default_docks(qapp, monkeypatch):
    monkeypatch.setattr(TvViewWindow, "_install_log_handler", lambda self: None)
    window = TvViewWindow()
    window.show()
    qapp.processEvents()
    toolbar = window.findChild(QtWidgets.QToolBar, "topovec.view.toolbar")

    assert window.windowTitle() == "topovec view"
    assert window.sources_dock.windowTitle() == "Data Sources"
    assert window.properties_dock.windowTitle() == "Properties"
    assert window.snippet_dock.windowTitle() == "Python Snippet"
    assert window.log_dock.windowTitle() == "Log"
    assert (
        window.dockWidgetArea(window.snippet_dock)
        == QtCore.Qt.DockWidgetArea.BottomDockWidgetArea
    )
    assert (
        window.dockWidgetArea(window.log_dock)
        == QtCore.Qt.DockWidgetArea.BottomDockWidgetArea
    )
    assert window.scene_combo.isEnabled() is False
    assert isinstance(window.viewport, QtViewport)
    assert window.viewport.testAttribute(QtCore.Qt.WidgetAttribute.WA_OpaquePaintEvent)
    assert window.viewport.autoFillBackground() is False
    assert window.sources_dock.isHidden() is False
    assert window.properties_dock.isHidden() is False
    assert window.snippet_dock.isHidden() is True
    assert window.log_dock.isHidden() is True
    assert window.left_panel_action.isChecked() is True
    assert window.bottom_panel_action.isChecked() is False
    assert window.right_panel_action.isChecked() is True
    assert toolbar is not None
    assert toolbar.toolButtonStyle() == QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly
    assert window.open_action.icon().isNull() is False
    assert window.save_image_action.icon().isNull() is False
    assert window.reset_camera_action.icon().isNull() is False
    assert window.snap_camera_action.icon().isNull() is False
    assert window.toggle_projection_action.icon().isNull() is False
    assert set(window._theme_actions) == {"system", "dark", "light"}
    assert window._theme_actions["system"].isChecked() is True
    window.close()


def test_toolbar_icon_helper_prefers_theme_and_falls_back_to_qstyle(qapp, monkeypatch):
    window = TvViewWindow()
    fallback_icon = QtGui.QIcon()
    fallback_icon.addPixmap(QtGui.QPixmap(16, 16))
    theme_icon = QtGui.QIcon()
    theme_icon.addPixmap(QtGui.QPixmap(16, 16))

    class DummyStyle:
        def standardIcon(self, _pixmap):
            return fallback_icon

    monkeypatch.setattr(TvViewWindow, "style", lambda self: DummyStyle())

    monkeypatch.setattr(
        QtGui.QIcon,
        "fromTheme",
        staticmethod(lambda name: theme_icon if name == "document-open" else QtGui.QIcon()),
    )

    selected = window._icon_from_theme(
        ("document-open", "folder-open"),
        QtWidgets.QStyle.StandardPixmap.SP_DialogOpenButton,
    )
    fallback_selected = window._icon_from_theme(
        ("missing-icon",),
        QtWidgets.QStyle.StandardPixmap.SP_DialogOpenButton,
    )

    assert selected.cacheKey() == theme_icon.cacheKey()
    assert fallback_selected.cacheKey() == fallback_icon.cacheKey()
    window.close()


def test_toolbar_icon_support_bootstraps_system_theme_paths(monkeypatch, tmp_path):
    icons_root = tmp_path / "icons"
    theme_dir = icons_root / "Adwaita"
    theme_dir.mkdir(parents=True)
    (theme_dir / "index.theme").write_text("[Icon Theme]\nName=Adwaita\n", encoding="utf-8")

    state = {
        "search_paths": [":/icons"],
        "theme_name": "",
        "fallback_theme_name": "",
    }

    monkeypatch.setenv("XDG_DATA_DIRS", str(tmp_path))
    monkeypatch.setattr(QtGui.QIcon, "themeSearchPaths", staticmethod(lambda: list(state["search_paths"])))
    monkeypatch.setattr(
        QtGui.QIcon,
        "setThemeSearchPaths",
        staticmethod(lambda paths: state.__setitem__("search_paths", list(paths))),
    )
    monkeypatch.setattr(QtGui.QIcon, "themeName", staticmethod(lambda: state["theme_name"]))
    monkeypatch.setattr(
        QtGui.QIcon,
        "fallbackThemeName",
        staticmethod(lambda: state["fallback_theme_name"]),
    )
    monkeypatch.setattr(
        QtGui.QIcon,
        "setFallbackThemeName",
        staticmethod(lambda name: state.__setitem__("fallback_theme_name", str(name))),
    )

    TvViewWindow._ensure_theme_icon_support()

    assert str(icons_root) in state["search_paths"]
    assert state["fallback_theme_name"] == "Adwaita"


def test_toolbar_separates_camera_actions_from_other_icons(qapp):
    window = TvViewWindow()
    toolbar = window.findChild(QtWidgets.QToolBar, "topovec.view.toolbar")
    actions = toolbar.actions()

    assert actions[0] is window.open_action
    assert actions[1] is window.save_image_action
    assert actions[2].isSeparator() is True
    assert actions[3] is window.left_panel_action
    assert actions[4] is window.bottom_panel_action
    assert actions[5] is window.right_panel_action
    assert actions[6].isSeparator() is True
    assert actions[7] is window.reset_camera_action
    assert actions[8] is window.snap_camera_action
    assert actions[9] is window.toggle_projection_action
    assert actions[10].isSeparator() is False
    assert toolbar.widgetForAction(actions[12]) is window.scene_combo
    window.close()


def test_view_menu_contains_area_actions_above_per_dock_toggles(qapp):
    window = TvViewWindow()
    view_menu = _menu_for_title(window, "&View")
    action_texts = [action.text() for action in view_menu.actions() if not action.isSeparator()]

    left_index = action_texts.index("Left Panel")
    bottom_index = action_texts.index("Bottom Panel")
    right_index = action_texts.index("Right Panel")
    sources_index = action_texts.index(window.sources_dock.toggleViewAction().text())
    properties_index = action_texts.index(window.properties_dock.toggleViewAction().text())
    snippet_index = action_texts.index(window.snippet_dock.toggleViewAction().text())
    log_index = action_texts.index(window.log_dock.toggleViewAction().text())

    assert left_index < sources_index
    assert bottom_index < snippet_index
    assert right_index < properties_index
    assert log_index > snippet_index
    window.close()


def test_default_dock_sizes_use_decluttered_widths(qapp, monkeypatch):
    window = TvViewWindow()
    calls = []

    monkeypatch.setattr(
        window,
        "resizeDocks",
        lambda docks, sizes, orientation: calls.append((list(docks), list(sizes), orientation)),
    )

    window._apply_default_dock_sizes()

    assert calls == [
        (
            [window.sources_dock, window.properties_dock],
            [240, 320],
            QtCore.Qt.Orientation.Horizontal,
        )
    ]
    window.close()


def test_reset_layout_restores_decluttered_default_docks(qapp, monkeypatch):
    monkeypatch.setattr(TvViewWindow, "_install_log_handler", lambda self: None)
    window = TvViewWindow()
    window.show()
    qapp.processEvents()

    window.bottom_panel_action.trigger()
    qapp.processEvents()
    window.left_panel_action.trigger()
    qapp.processEvents()

    assert window.snippet_dock.isHidden() is False
    assert window.log_dock.isHidden() is False
    assert window.sources_dock.isHidden() is True

    window.reset_layout()
    qapp.processEvents()

    assert window.sources_dock.isHidden() is False
    assert window.properties_dock.isHidden() is False
    assert window.snippet_dock.isHidden() is True
    assert window.log_dock.isHidden() is True
    assert window.left_panel_action.isChecked() is True
    assert window.bottom_panel_action.isChecked() is False
    assert window.right_panel_action.isChecked() is True
    window.close()


def test_area_actions_toggle_corresponding_dock_areas(qapp, monkeypatch):
    monkeypatch.setattr(TvViewWindow, "_install_log_handler", lambda self: None)
    window = TvViewWindow()
    window.show()
    qapp.processEvents()

    window.left_panel_action.trigger()
    qapp.processEvents()
    assert window.sources_dock.isHidden() is True
    assert window.left_panel_action.isChecked() is False

    window.left_panel_action.trigger()
    qapp.processEvents()
    assert window.sources_dock.isHidden() is False
    assert window.left_panel_action.isChecked() is True

    window.right_panel_action.trigger()
    qapp.processEvents()
    assert window.properties_dock.isHidden() is True
    assert window.right_panel_action.isChecked() is False

    window.right_panel_action.trigger()
    qapp.processEvents()
    assert window.properties_dock.isHidden() is False
    assert window.right_panel_action.isChecked() is True

    window.bottom_panel_action.trigger()
    qapp.processEvents()
    assert window.snippet_dock.isHidden() is False
    assert window.log_dock.isHidden() is False
    assert window.bottom_panel_action.isChecked() is True

    window.bottom_panel_action.trigger()
    qapp.processEvents()
    assert window.snippet_dock.isHidden() is True
    assert window.log_dock.isHidden() is True
    assert window.bottom_panel_action.isChecked() is False
    window.close()


def test_bottom_area_action_applies_to_additional_bottom_docks(qapp, monkeypatch):
    monkeypatch.setattr(TvViewWindow, "_install_log_handler", lambda self: None)
    window = TvViewWindow()
    extra_dock = QtWidgets.QDockWidget("Extra", window)
    extra_dock.setObjectName("topovec.view.extra")
    extra_dock.setWidget(QtWidgets.QLabel("extra"))
    window.addDockWidget(QtCore.Qt.DockWidgetArea.BottomDockWidgetArea, extra_dock)
    window.show()
    qapp.processEvents()

    assert window.bottom_panel_action.isChecked() is True

    window.bottom_panel_action.trigger()
    qapp.processEvents()

    assert window.snippet_dock.isHidden() is True
    assert window.log_dock.isHidden() is True
    assert extra_dock.isHidden() is True
    assert window.bottom_panel_action.isChecked() is False

    window.bottom_panel_action.trigger()
    qapp.processEvents()

    assert window.snippet_dock.isHidden() is False
    assert window.log_dock.isHidden() is False
    assert extra_dock.isHidden() is False
    assert window.bottom_panel_action.isChecked() is True
    window.close()


def test_dock_area_actions_track_manual_visibility_location_and_floating(qapp):
    window = TvViewWindow()
    window.show()
    qapp.processEvents()

    window.sources_dock.hide()
    qapp.processEvents()
    assert window.left_panel_action.isChecked() is False

    window.sources_dock.show()
    qapp.processEvents()
    assert window.left_panel_action.isChecked() is True

    window.properties_dock.setFloating(True)
    qapp.processEvents()
    assert window.right_panel_action.isChecked() is False

    window.properties_dock.setFloating(False)
    window.addDockWidget(QtCore.Qt.DockWidgetArea.LeftDockWidgetArea, window.properties_dock)
    qapp.processEvents()
    assert window.right_panel_action.isChecked() is False
    assert window.left_panel_action.isChecked() is True
    window.close()


def test_theme_selection_persists_and_restores(qapp, topovec_view_settings_path):
    window = TvViewWindow()
    initial_window = qapp.palette().color(QtGui.QPalette.ColorRole.Window)

    window._theme_actions["dark"].trigger()

    assert _read_theme_setting(topovec_view_settings_path) == "dark"
    assert window._theme_actions["dark"].isChecked() is True
    assert qapp.palette().color(QtGui.QPalette.ColorRole.Window) != initial_window
    window.close()

    restored = TvViewWindow()
    assert restored._theme_actions["dark"].isChecked() is True
    assert _read_theme_setting(topovec_view_settings_path) == "dark"
    restored.close()


def test_theme_modes_apply_distinct_palettes(qapp, topovec_view_settings_path):
    window = TvViewWindow()
    initial = qapp.style().standardPalette().color(QtGui.QPalette.ColorRole.Window)

    window._theme_actions["dark"].trigger()
    dark = qapp.palette().color(QtGui.QPalette.ColorRole.Window)
    window._theme_actions["light"].trigger()
    light = qapp.palette().color(QtGui.QPalette.ColorRole.Window)
    window._theme_actions["system"].trigger()
    system = qapp.palette().color(QtGui.QPalette.ColorRole.Window)

    assert dark != light
    assert system == qapp.style().standardPalette().color(QtGui.QPalette.ColorRole.Window)
    assert dark != initial or light != initial
    window.close()


def test_system_theme_change_only_reapplies_in_system_mode(qapp, monkeypatch):
    window = TvViewWindow()
    calls = []
    original_apply_theme = window._apply_theme

    def tracked_apply_theme(mode, *, persist):
        calls.append((mode, persist))
        return original_apply_theme(mode, persist=persist)

    monkeypatch.setattr(window, "_apply_theme", tracked_apply_theme)

    window._theme_actions["dark"].trigger()
    calls.clear()
    window._on_system_color_scheme_changed(QtCore.Qt.ColorScheme.Light)
    assert calls == []

    window._theme_actions["system"].trigger()
    calls.clear()
    window._on_system_color_scheme_changed(QtCore.Qt.ColorScheme.Dark)
    assert calls == [("system", False)]
    window.close()


def test_property_panel_updates_theme_sensitive_reset_icon(qapp):
    panel = PropertyPanel()
    panel.set_snapshot(
        {
            "property_groups": [
                {
                    "id": "Render",
                    "title": "Render",
                    "properties": [
                        {
                            "index": 0,
                            "kind": "numeric",
                            "title": "Level",
                            "field": "level",
                            "value": 0.75,
                            "min": 0.0,
                            "max": 1.0,
                            "count": 5,
                            "step": 0.25,
                            "input_kind": "number",
                            "row_kind": "inline",
                            "default_value": 0.5,
                            "is_default": False,
                            "can_reset": True,
                        }
                    ],
                }
            ]
        }
    )

    reset = next(
        button
        for button in panel.findChildren(QtWidgets.QToolButton)
        if button.property("inspectorRole") == "reset"
    )
    before = reset.icon().cacheKey()

    from topovec.viewer.qt.theme import theme_spec_for_mode

    panel.apply_theme(theme_spec_for_mode("dark", qapp))
    after = reset.icon().cacheKey()

    assert before != after


def test_viewport_empty_clear_color_follows_theme(qapp):
    viewport = QtViewportWindow()
    from topovec.viewer.qt.theme import theme_spec_for_mode

    viewport.apply_theme(theme_spec_for_mode("dark", qapp))
    dark = viewport._empty_clear_color()
    viewport.apply_theme(theme_spec_for_mode("light", qapp))
    light = viewport._empty_clear_color()

    assert dark != light
    assert dark == viewport._theme_spec.viewport_clear_rgba or light == viewport._theme_spec.viewport_clear_rgba


def test_load_paths_updates_recent_files_mru_and_caps_history(
    qapp, monkeypatch, topovec_view_settings_path
):
    window = TvViewWindow()
    documents = {}
    paths = []
    for index in range(11):
        path = Path(f"/tmp/recent_{index}.npz")
        paths.append(path)
        path.touch()
        documents[str(path.resolve())] = _make_document(
            f"recent_{index}.npz",
            System.cubic((4, 4, 4)),
            [np.zeros((4, 4, 4, 1, 3), dtype=np.float32)],
            format_name="NPZ",
        )
    monkeypatch.setattr(
        "topovec.viewer.qt.main_window.load_viewer_document",
        lambda raw_path: documents[str(Path(raw_path).expanduser().resolve())],
    )

    window.load_paths([str(path) for path in paths])
    qapp.processEvents()

    stored = _read_recent_settings_paths(topovec_view_settings_path)
    assert stored == [str(path.resolve()) for path in paths[:10]]

    window.load_paths([str(paths[5])])
    qapp.processEvents()

    stored = _read_recent_settings_paths(topovec_view_settings_path)
    assert stored[0] == str(paths[5].resolve())
    assert len(stored) == 10
    assert len(set(stored)) == 10
    window.close()


def test_recent_menu_rebuilds_from_persisted_settings_and_disambiguates_labels(
    qapp, topovec_view_settings_path
):
    path_a = (topovec_view_settings_path.parent / "dir_a" / "shared.npz").resolve()
    path_b = (topovec_view_settings_path.parent / "dir_b" / "shared.npz").resolve()
    path_a.parent.mkdir(parents=True, exist_ok=True)
    path_b.parent.mkdir(parents=True, exist_ok=True)
    path_a.touch()
    path_b.touch()
    settings = REAL_QSETTINGS(
        str(topovec_view_settings_path),
        REAL_QSETTINGS.Format.IniFormat,
    )
    settings.setValue("recent_files", [str(path_a), str(path_b)])

    window = TvViewWindow()
    window._refresh_recent_files_menu()

    labels = [action.text() for action in window.open_recent_menu.actions() if action.text()]

    assert labels[:2] == [f"shared.npz — {path_a.parent}", f"shared.npz — {path_b.parent}"]
    assert labels[-1] == "Clear Recent Files"
    assert window.open_recent_menu.menuAction().isEnabled() is True
    window.close()


def test_open_recent_file_focuses_existing_document_without_duplicate_source(
    qapp, monkeypatch, topovec_view_settings_path
):
    window = TvViewWindow()
    window.viewport._ctx = object()
    _install_qt_scene_registry(monkeypatch)

    document_a = _make_document(
        "field_a.sad",
        System.cubic((4, 4, 4)),
        [np.zeros((4, 4, 4, 1, 3), dtype=np.float32)],
    )
    document_b = _make_document(
        "field_b.sad",
        System.cubic((4, 4, 4)),
        [np.ones((4, 4, 4, 1, 3), dtype=np.float32)],
    )
    documents = {
        str(document_a.path): document_a,
        str(document_b.path): document_b,
    }
    document_a.path.touch()
    document_b.path.touch()
    monkeypatch.setattr(
        "topovec.viewer.qt.main_window.load_viewer_document",
        lambda raw_path: documents[str(Path(raw_path).expanduser().resolve())],
    )

    window.load_paths([str(document_a.path), str(document_b.path)])
    qapp.processEvents()
    second_item = window.sources_tree.topLevelItem(1).child(0)
    window.sources_tree.setCurrentItem(second_item)
    qapp.processEvents()

    window.open_recent_file(str(document_a.path))
    qapp.processEvents()

    current_payload = window.sources_tree.currentItem().data(0, QtCore.Qt.ItemDataRole.UserRole)
    assert window.sources_tree.topLevelItemCount() == 2
    assert current_payload[0] == "state"
    assert current_payload[1] is document_a
    assert window._current_document is document_a
    stored = _read_recent_settings_paths(topovec_view_settings_path)
    assert stored[0] == str(document_a.path)
    window.close()


def test_recent_menu_prunes_missing_paths_and_clear_empties_settings(
    qapp, topovec_view_settings_path
):
    existing = (topovec_view_settings_path.parent / "kept.npz").resolve()
    existing.touch()
    missing = (topovec_view_settings_path.parent / "missing.npz").resolve()
    settings = REAL_QSETTINGS(
        str(topovec_view_settings_path),
        REAL_QSETTINGS.Format.IniFormat,
    )
    settings.setValue("recent_files", [str(existing), str(missing)])

    window = TvViewWindow()
    window._refresh_recent_files_menu()

    stored = _read_recent_settings_paths(topovec_view_settings_path)
    labels = [action.text() for action in window.open_recent_menu.actions() if action.text()]
    assert stored == [str(existing)]
    assert labels == ["kept.npz", "Clear Recent Files"]

    window.clear_recent_files()

    assert _read_recent_settings_paths(topovec_view_settings_path) == []
    assert window.open_recent_menu.menuAction().isEnabled() is False
    window.close()


def test_closing_current_source_keeps_recent_history(
    qapp, monkeypatch, topovec_view_settings_path
):
    window = TvViewWindow()
    document = _make_document(
        "close_me.npz",
        System.cubic((4, 4, 4)),
        [np.zeros((4, 4, 4, 1, 3), dtype=np.float32)],
        format_name="NPZ",
    )
    document.path.touch()
    monkeypatch.setattr(
        "topovec.viewer.qt.main_window.load_viewer_document",
        lambda _raw_path: document,
    )

    window.load_paths([str(document.path)])
    qapp.processEvents()
    window.close_selected_source()
    qapp.processEvents()

    assert _read_recent_settings_paths(topovec_view_settings_path) == [str(document.path)]
    window.close()


def test_window_ignores_legacy_tvview_settings_namespace(qapp, monkeypatch, tmp_path):
    settings_path = tmp_path / "topovec.view.ini"
    legacy_path = tmp_path / "tvview.ini"
    legacy_state = (legacy_path.parent / "legacy_state.npz").resolve()
    legacy_state.touch()

    legacy_settings = REAL_QSETTINGS(str(legacy_path), REAL_QSETTINGS.Format.IniFormat)
    legacy_settings.setValue(THEME_SETTING_KEY, "dark")
    legacy_settings.setValue("recent_files", [str(legacy_state)])

    def _make_settings(organization, application):
        assert organization == "topovec"
        assert application == "topovec.view"
        return REAL_QSETTINGS(str(settings_path), REAL_QSETTINGS.Format.IniFormat)

    monkeypatch.setattr("topovec.viewer.qt.main_window.QtCore.QSettings", _make_settings)

    window = TvViewWindow()
    window._refresh_recent_files_menu()

    assert window._theme_actions["system"].isChecked() is True
    assert _read_theme_setting(settings_path) is None
    assert _read_recent_settings_paths(settings_path) == []
    assert window.open_recent_menu.menuAction().isEnabled() is False
    window.close()


def test_log_panel_supports_append_clear_and_copy(qapp):
    panel = LogPanel()

    panel.append_message("12:00:00 INFO topovec.viewer.qt: hello", logging.INFO)
    assert "hello" in panel.text()

    panel.copy_to_clipboard()
    assert QtWidgets.QApplication.clipboard().text() == panel.text()

    panel.clear()
    assert panel.text() == ""


def test_warning_log_records_raise_log_dock(qapp, monkeypatch):
    window = TvViewWindow()
    raised = []

    monkeypatch.setattr(window.log_dock, "raise_", lambda: raised.append("raised"))

    window._on_log_record_appended(logging.INFO)
    assert raised == []

    window._on_log_record_appended(logging.WARNING)
    assert raised == ["raised"]
    assert window.log_dock.isHidden() is False
    assert window.bottom_panel_action.isChecked() is True
    window.close()


def test_bottom_output_docks_default_to_log_tab(qapp, monkeypatch):
    snippet_raised = []
    log_raised = []

    monkeypatch.setattr(
        TvViewWindow,
        "_install_log_handler",
        lambda self: None,
    )
    monkeypatch.setattr(
        QtWidgets.QDockWidget,
        "raise_",
        lambda dock: (
            log_raised.append(dock.objectName())
            if dock.objectName() == "topovec.view.log"
            else snippet_raised.append(dock.objectName())
        ),
    )

    window = TvViewWindow()

    assert log_raised
    assert snippet_raised == []
    window.close()


def test_activation_failure_is_logged_and_reported(qapp, monkeypatch):
    window = TvViewWindow()
    window.viewport._ctx = object()

    document = ViewerDocument(
        path=Path("/tmp/example.npz"),
        label="example.npz",
        format_name="NPZ",
        system=SimpleNamespace(size=np.asarray((4, 4, 4), dtype=int)),
        entries=[],
        metadata={},
    )
    entry = StateEntry(
        label="PATH[0]",
        index=(0,),
        load_state=lambda: np.zeros((4, 4, 4, 1, 3), dtype=np.float32),
        metadata={},
    )

    monkeypatch.setattr(window, "_preferred_scene_for_system", lambda _system: "Layer")
    monkeypatch.setattr(
        window.viewport.host,
        "create_scene",
        lambda _scene_cls, _system: (_ for _ in ()).throw(RuntimeError("boom")),
    )

    window._activate_entry(document, entry)
    qapp.processEvents()

    assert "Failed to activate" in window.statusBar().currentMessage()
    assert "boom" in window.log_panel.text()
    window.close()


def test_load_failure_is_logged_and_reported(qapp, monkeypatch):
    window = TvViewWindow()
    warnings = []

    monkeypatch.setattr(
        "topovec.viewer.qt.main_window.load_viewer_document",
        lambda _path: (_ for _ in ()).throw(ValueError("broken file")),
    )
    monkeypatch.setattr(
        QtWidgets.QMessageBox,
        "warning",
        lambda *args: warnings.append(args),
    )

    window.load_paths(["/tmp/bad.npz"])
    qapp.processEvents()

    assert warnings
    assert "Failed to open bad.npz" in window.statusBar().currentMessage()
    assert "broken file" in window.log_panel.text()
    window.close()


def test_qt_opengl_loader_uses_get_proc_address():
    calls = []

    class DummyContext:
        def getProcAddress(self, name):
            calls.append(name)
            return 12345

    loader = QtOpenGLLoader(DummyContext())

    assert loader.load_opengl_function("glExample") == 12345
    assert calls == ["glExample"]


def test_initialize_gl_prefers_qt_loader(qapp, monkeypatch):
    viewport = QtViewportWindow()
    dummy_ctx = object()
    called = {}

    class DummyQtContext:
        def getProcAddress(self, name):
            called.setdefault("proc_names", []).append(name)
            return 1

    monkeypatch.setattr(QtGui.QOpenGLContext, "currentContext", staticmethod(lambda: DummyQtContext()))
    monkeypatch.setattr(viewport, "context", lambda: DummyQtContext())

    def fake_init_context(loader):
        called["loader_type"] = type(loader).__name__
        called["proc"] = loader.load_opengl_function("glGetString")

    monkeypatch.setattr("topovec.viewer.qt.viewport.mgl.init_context", fake_init_context)
    monkeypatch.setattr("topovec.viewer.qt.viewport.mgl.get_context", lambda: dummy_ctx)
    monkeypatch.setattr("topovec.viewer.qt.viewport.mgl.create_context", lambda: (_ for _ in ()).throw(AssertionError("fallback should not run")))

    viewport.initializeGL()

    assert viewport._ctx is dummy_ctx
    assert called["loader_type"] == "QtOpenGLLoader"
    assert called["proc"] == 1


def test_initialize_gl_falls_back_to_default_loader(qapp, monkeypatch):
    viewport = QtViewportWindow()
    fallback_ctx = object()

    class DummyQtContext:
        def getProcAddress(self, name):
            return 1

    messages = []

    monkeypatch.setattr(QtGui.QOpenGLContext, "currentContext", staticmethod(lambda: DummyQtContext()))
    monkeypatch.setattr(viewport, "context", lambda: DummyQtContext())
    monkeypatch.setattr(
        "topovec.viewer.qt.viewport.mgl.init_context",
        lambda loader: (_ for _ in ()).throw(RuntimeError("qt loader failed")),
    )
    monkeypatch.setattr("topovec.viewer.qt.viewport.mgl.create_context", lambda: fallback_ctx)
    viewport.runtime_error.connect(messages.append)

    viewport.initializeGL()

    assert viewport._ctx is fallback_ctx
    assert messages == []


def test_current_widget_target_uses_detected_default_framebuffer(monkeypatch):
    viewport = QtViewportWindow()
    detected = []
    detected_fbo = SimpleNamespace(glo=23, size=(320, 240))

    class DummyContext:
        def detect_framebuffer(self, framebuffer_id):
            detected.append(framebuffer_id)
            return detected_fbo

    monkeypatch.setattr(viewport, "defaultFramebufferObject", lambda: 23)
    monkeypatch.setattr(viewport, "current_render_size", lambda: (320, 240))
    viewport._ctx = DummyContext()

    target = viewport._current_widget_target()

    assert detected == [23]
    assert target.fbo is detected_fbo
    assert target.size == (320, 240)


def test_prepare_render_target_restores_known_baseline():
    viewport = QtViewportWindow()

    class DummyFramebuffer:
        def __init__(self):
            self.glo = 77
            self.size = (320, 240)
            self.color_mask = (False, False, False, False)
            self.depth_mask = False
            self.viewport = (1, 2, 3, 4)
            self.scissor = (5, 6, 7, 8)

    class DummyContext:
        def __init__(self):
            self.fbo = DummyFramebuffer()
            self._blend_func = (1, 1)
            self._blend_equation = 12345
            self.enabled = []
            self.viewport = None
            self._scissor = (5, 6, 7, 8)

        def enable_only(self, flags):
            self.enabled.append(flags)

        @property
        def scissor(self):
            return self._scissor

        @scissor.setter
        def scissor(self, value):
            self._scissor = value

        @property
        def blend_func(self):
            raise NotImplementedError()

        @blend_func.setter
        def blend_func(self, value):
            self._blend_func = tuple(value)

        @property
        def blend_equation(self):
            raise NotImplementedError()

        @blend_equation.setter
        def blend_equation(self, value):
            self._blend_equation = value

    viewport._ctx = DummyContext()
    bind_calls = []

    class DummyTarget:
        size = (320, 240)

        def __init__(self):
            self.fbo = viewport._ctx.fbo

        def bind(self):
            bind_calls.append("bind")
            return self

    viewport._prepare_render_target("paintGL", target=DummyTarget())

    assert bind_calls == ["bind"]
    assert viewport._ctx.fbo.color_mask == (True, True, True, True)
    assert viewport._ctx.fbo.depth_mask is True
    assert viewport._ctx.fbo.scissor is None
    assert viewport._ctx.fbo.viewport == (0, 0, 320, 240)
    assert viewport._ctx.viewport == (0, 0, 320, 240)
    assert viewport._ctx.enabled == [mgl.NOTHING]
    assert viewport._ctx.scissor is None
    assert viewport._ctx._blend_func == mgl.DEFAULT_BLENDING
    assert viewport._ctx._blend_equation == mgl.FUNC_ADD


def test_runtime_diagnostics_collects_gl_and_qt_context_details():
    viewport = QtViewport()
    viewport._ctx = SimpleNamespace(
        info={
            "GL_VENDOR": "Vendor",
            "GL_RENDERER": "Renderer",
            "GL_VERSION": "4.6",
            "GL_SHADING_LANGUAGE_VERSION": "4.60",
        },
        version_code=460,
    )

    monkey_format = QtGui.QSurfaceFormat()
    monkey_format.setVersion(4, 1)
    monkey_format.setProfile(QtGui.QSurfaceFormat.OpenGLContextProfile.CoreProfile)
    monkey_format.setRenderableType(QtGui.QSurfaceFormat.RenderableType.OpenGL)
    monkey_format.setSwapInterval(1)
    monkey_format.setSamples(4)
    monkey_format.setDepthBufferSize(24)
    monkey_format.setStencilBufferSize(8)
    monkey_format.setAlphaBufferSize(8)

    class DummyQtContext:
        def format(self):
            return monkey_format

        def isOpenGLES(self):
            return False

    viewport.setFormat(monkey_format)
    diagnostics = viewport._runtime_diagnostics(DummyQtContext())

    assert diagnostics["vendor"] == "Vendor"
    assert diagnostics["renderer"] == "Renderer"
    assert diagnostics["version"] == "4.6"
    assert diagnostics["glsl"] == "4.60"
    assert diagnostics["version_code"] == 460
    assert diagnostics["is_opengles"] is False
    assert diagnostics["actual_format"]["version"] == (4, 1)
    assert diagnostics["requested_format"]["version"] == (4, 1)


def test_paint_gl_without_scene_clears_current_framebuffer(monkeypatch):
    viewport = QtViewportWindow()
    finish_calls = []
    clear_calls = []
    prepared = []

    class DummyContext:
        def finish(self):
            finish_calls.append("finish")

    viewport._ctx = DummyContext()

    class DummyTarget:
        size = (320, 240)
        fbo = SimpleNamespace(glo=77)
        def clear_color(self, *args):
            clear_calls.append(args)

    target = DummyTarget()
    monkeypatch.setattr(viewport, "_current_widget_target", lambda: target)
    monkeypatch.setattr(
        viewport,
        "_prepare_render_target",
        lambda reason, *, target: prepared.append((reason, target)),
    )

    viewport.paintGL()

    assert clear_calls == [viewport._empty_clear_color()]
    assert prepared == [("paintGL", target)]
    assert finish_calls == ["finish"]


def test_render_image_uses_explicit_render_target(monkeypatch):
    viewport = QtViewport()
    make_current_calls = []
    done_current_calls = []
    prepared = []
    release_calls = []

    class DummyContext:
        def finish(self):
            return None

    class DummyTarget:
        def __init__(self, ctx, size):
            self.ctx = ctx
            self.size = tuple(size)
            self.fbo = SimpleNamespace(glo=77)
            self.color_texture = SimpleNamespace(glo=15)

        def bind(self):
            return self

        def read_rgb(self):
            width, height = self.size
            return b"\x00" * (width * height * 3)

        def release(self):
            release_calls.append(self.size)

    class DummyScene:
        def __init__(self):
            self.targets = []

        def render_to(self, target):
            self.targets.append(target)

    monkeypatch.setattr("topovec.viewer.qt.viewport.OwnedRenderTarget", DummyTarget)
    monkeypatch.setattr(viewport, "makeCurrent", lambda: make_current_calls.append("makeCurrent"))
    monkeypatch.setattr(viewport, "doneCurrent", lambda: done_current_calls.append("doneCurrent"))
    monkeypatch.setattr(
        viewport,
        "_prepare_render_target",
        lambda reason, *, target: prepared.append((reason, target.size)),
    )
    viewport._ctx = DummyContext()
    viewport._scene = DummyScene()

    image = viewport.host.render_image(size=(64, 48))

    assert image.size == (64, 48)
    assert prepared == [("render_image", (64, 48))]
    assert len(viewport._scene.targets) == 1
    assert viewport._scene.targets[0].size == (64, 48)
    assert release_calls == [(64, 48)]
    assert make_current_calls == ["makeCurrent"]
    assert done_current_calls == ["doneCurrent"]


def test_camera_session_state_change_does_not_refresh_properties_panel(qapp, monkeypatch):
    window = TvViewWindow()
    property_calls = []
    summary_calls = []

    class DummySession:
        selected_scene = "Layer"

        def camera_snapshot(self):
            return {"projection_label": "Perspective"}

        def render_snippet(self):
            return "snippet"

    monkeypatch.setattr(window.properties_panel, "set_snapshot", lambda *args, **kwargs: property_calls.append((args, kwargs)))
    monkeypatch.setattr(window.snippet_dock, "isVisible", lambda: True)
    monkeypatch.setattr(window.snippet_panel, "set_text", lambda text: summary_calls.append(text))
    window._session = DummySession()
    window._current_document = SimpleNamespace(label="doc.sad")
    window._current_entry = SimpleNamespace(label="PATH[0]")

    window._on_session_state_changed(False)
    window._on_session_state_changed(True)

    assert property_calls == []
    assert summary_calls == ["snippet"]
    window.close()


def test_qt_viewport_host_scene_mutation_scope_balances_context(monkeypatch):
    viewport = QtViewportWindow()
    calls = []

    monkeypatch.setattr(viewport, "makeCurrent", lambda: calls.append("makeCurrent"))
    monkeypatch.setattr(viewport, "doneCurrent", lambda: calls.append("doneCurrent"))

    with viewport.host.scene_mutation_scope("apply_property_change"):
        calls.append("body")

    assert calls == ["makeCurrent", "body", "doneCurrent"]


def test_qt_viewport_host_scene_mutation_scope_unwinds_after_failure(monkeypatch):
    viewport = QtViewportWindow()
    calls = []

    monkeypatch.setattr(viewport, "makeCurrent", lambda: calls.append("makeCurrent"))
    monkeypatch.setattr(viewport, "doneCurrent", lambda: calls.append("doneCurrent"))

    with pytest.raises(RuntimeError, match="boom"):
        with viewport.host.scene_mutation_scope("apply_property_change"):
            calls.append("body")
            raise RuntimeError("boom")

    assert calls == ["makeCurrent", "body", "doneCurrent"]


def test_attach_upload_and_set_session_do_not_trigger_implicit_redraws(monkeypatch):
    viewport = QtViewport()
    update_calls = []

    class DummyScene:
        def __init__(self):
            self.uploaded = []

        def upload(self, state):
            self.uploaded.append(state)

    monkeypatch.setattr(viewport, "update", lambda: update_calls.append("update"))
    monkeypatch.setattr(viewport, "makeCurrent", lambda: None)
    monkeypatch.setattr(viewport, "doneCurrent", lambda: None)
    viewport._ctx = object()
    scene = DummyScene()

    viewport.host.attach_scene(scene)
    viewport.host.upload(np.asarray([1.0], dtype=np.float32))
    viewport.set_session(object())

    assert len(scene.uploaded) == 1
    assert np.allclose(scene.uploaded[0], np.asarray([1.0], dtype=np.float32))
    assert update_calls == []


def test_refresh_ui_skips_snippet_generation_when_snippet_tab_hidden(qapp, monkeypatch):
    window = TvViewWindow()
    snippet_calls = []

    class DummySession:
        available_scenes = ["Layer"]
        selected_scene = "Layer"

        def properties_snapshot(self):
            return {"property_groups": [], "selected_scene": "Layer", "available_scenes": ["Layer"]}

        def render_snippet(self):
            snippet_calls.append("called")
            return "snippet"

        def camera_snapshot(self):
            return {"projection_label": "Perspective"}

    monkeypatch.setattr(window.snippet_dock, "isVisible", lambda: False)
    window._session = DummySession()
    window._current_document = SimpleNamespace(label="doc.npz")
    window._current_entry = SimpleNamespace(label="PATH[0]")

    window._refresh_ui(full=True)

    assert snippet_calls == []
    window.close()


def test_activation_with_visible_snippet_does_not_create_extra_scene(qapp, monkeypatch):
    window = TvViewWindow()
    window.viewport._ctx = object()
    create_calls = []
    update_calls = []

    class DummyScene:
        def __init__(self, mglctx=None, system=None):
            self._mglctx = mglctx
            self._system = system
            self.camera = Camera()
            self.uploaded_state = None

        @property
        def system(self):
            return self._system

        @classmethod
        def supports_system(cls, system):
            return True

        def upload(self, state=None):
            self.uploaded_state = state

        def list_properties(self):
            return {"Render": []}

    registry = {"Layer": DummyScene}
    monkeypatch.setattr("topovec.viewer.qt.main_window.SCENES", dict(registry))
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: dict(registry),
    )
    original_create_scene = window.viewport.host.create_scene

    def counted_create_scene(scene_cls, system):
        create_calls.append(getattr(scene_cls, "__name__", repr(scene_cls)))
        return original_create_scene(scene_cls, system)

    monkeypatch.setattr(window.viewport.host, "create_scene", counted_create_scene)
    monkeypatch.setattr(window.snippet_dock, "isVisible", lambda: True)
    monkeypatch.setattr(window.viewport, "update", lambda: update_calls.append("update"))

    document = ViewerDocument(
        path=Path("/tmp/example.npz"),
        label="example.npz",
        format_name="NPZ",
        system=SimpleNamespace(size=np.asarray((4, 4, 4), dtype=int)),
        entries=[],
        metadata={},
    )
    entry = StateEntry(
        label="PATH[0]",
        index=(0,),
        load_state=lambda: np.zeros((4, 4, 4, 1, 3), dtype=np.float32),
        metadata={},
    )

    window._activate_entry(document, entry)
    qapp.processEvents()

    assert create_calls == ["DummyScene"]
    assert update_calls == ["update"]
    assert "render_prepare('Layer', system)" in window.snippet_panel._editor.toPlainText()
    window.close()


def test_same_system_activation_reuses_session_and_preserves_state(qapp, monkeypatch):
    window = TvViewWindow()
    window.viewport._ctx = object()
    create_calls = []

    _install_qt_scene_registry(monkeypatch)
    original_create_scene = window.viewport.host.create_scene

    def counted_create_scene(scene_cls, system):
        create_calls.append(getattr(scene_cls, "__name__", repr(scene_cls)))
        return original_create_scene(scene_cls, system)

    monkeypatch.setattr(window.viewport.host, "create_scene", counted_create_scene)

    system = System.cubic((4, 4, 4))
    document = _make_document(
        "field.sad",
        system,
        [
            np.zeros((4, 4, 4, 1, 3), dtype=np.float32),
            np.ones((4, 4, 4, 1, 3), dtype=np.float32),
        ],
    )

    window._activate_entry(document, document.entries[0])
    qapp.processEvents()

    session = window.session
    assert session is not None
    assert session.switch_scene("Ball") is True
    session.apply_property_change(
        "Render",
        0,
        True,
        expected_field="flag",
        expected_title="Flag",
    )
    session.camera.look_scale *= 1.75
    session.commit_camera_interaction()
    camera_before = camera_state_signature(session.camera_snapshot())

    window._activate_entry(document, document.entries[1])
    qapp.processEvents()

    assert window.session is session
    assert create_calls == ["_QtLayerScene", "_QtBallScene"]
    assert window.session.selected_scene == "Ball"
    assert camera_state_signature(window.session.camera_snapshot()) == camera_before
    props = window.session.properties_snapshot()["property_groups"][0]["properties"]
    assert props[0]["value"] is True
    window.close()


def test_same_system_activation_reuses_session_across_documents(qapp, monkeypatch):
    window = TvViewWindow()
    window.viewport._ctx = object()
    create_calls = []

    _install_qt_scene_registry(monkeypatch)
    original_create_scene = window.viewport.host.create_scene

    def counted_create_scene(scene_cls, system):
        create_calls.append(getattr(scene_cls, "__name__", repr(scene_cls)))
        return original_create_scene(scene_cls, system)

    monkeypatch.setattr(window.viewport.host, "create_scene", counted_create_scene)

    document_a = _make_document(
        "field_a.sad",
        System.cubic((4, 4, 4)),
        [np.zeros((4, 4, 4, 1, 3), dtype=np.float32)],
    )
    document_b = _make_document(
        "field_b.sad",
        System.cubic((4, 4, 4)),
        [np.ones((4, 4, 4, 1, 3), dtype=np.float32)],
    )

    window._activate_entry(document_a, document_a.entries[0])
    qapp.processEvents()
    session = window.session
    assert session is not None
    session.camera.look_scale *= 1.5
    session.commit_camera_interaction()
    camera_before = camera_state_signature(session.camera_snapshot())

    window._activate_entry(document_b, document_b.entries[0])
    qapp.processEvents()

    assert window.session is session
    assert create_calls == ["_QtLayerScene"]
    assert camera_state_signature(window.session.camera_snapshot()) == camera_before
    window.close()


def test_qt_property_edit_runs_under_scene_mutation_scope(qapp, monkeypatch):
    window = TvViewWindow()
    window.viewport._ctx = object()
    mutation_state = {"active": False}
    original_make_current = window.viewport.makeCurrent
    original_done_current = window.viewport.doneCurrent
    calls = []

    monkeypatch.setattr(
        "topovec.viewer.qt.main_window.SCENES",
        {"Layer": _QtScopedVectorScene},
    )
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: {"Layer": _QtScopedVectorScene},
    )

    def tracked_make_current():
        mutation_state["active"] = True
        calls.append("makeCurrent")
        return original_make_current()

    def tracked_done_current():
        calls.append("doneCurrent")
        try:
            return original_done_current()
        finally:
            mutation_state["active"] = False

    monkeypatch.setattr(window.viewport, "makeCurrent", tracked_make_current)
    monkeypatch.setattr(window.viewport, "doneCurrent", tracked_done_current)

    original_create_scene = window.viewport.host.create_scene

    def counted_create_scene(scene_cls, system):
        scene = original_create_scene(scene_cls, system)
        scene._mutation_state = mutation_state
        return scene

    monkeypatch.setattr(window.viewport.host, "create_scene", counted_create_scene)

    document = _make_document(
        "field.sad",
        System.cubic((4, 4, 4)),
        [np.zeros((4, 4, 4, 1, 3), dtype=np.float32)],
    )

    window._activate_entry(document, document.entries[0])
    qapp.processEvents()

    window._on_property_edited("Vector", 0, "Arrow", "mesh_shape", "Shape")
    qapp.processEvents()

    assert window.session is not None
    assert window.session.scene.mesh_shape == "Arrow"
    assert (
        window.session.properties_snapshot()["property_groups"][0]["properties"][0][
            "value"
        ]
        == "Arrow"
    )
    assert calls[-2:] == ["makeCurrent", "doneCurrent"]
    window.close()


def test_qt_non_structural_property_edit_updates_panel_without_rebuild(
    qapp, monkeypatch
):
    window = TvViewWindow()
    window.viewport._ctx = object()
    property_calls = []
    mutation_state = {"active": False}
    original_make_current = window.viewport.makeCurrent
    original_done_current = window.viewport.doneCurrent

    monkeypatch.setattr(
        "topovec.viewer.qt.main_window.SCENES",
        {"Layer": _QtScopedVectorScene},
    )
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: {"Layer": _QtScopedVectorScene},
    )

    original_set_snapshot = window.properties_panel.set_snapshot

    def tracked_set_snapshot(snapshot, *, rebuild=True):
        property_calls.append(rebuild)
        return original_set_snapshot(snapshot, rebuild=rebuild)

    def tracked_make_current():
        mutation_state["active"] = True
        return original_make_current()

    def tracked_done_current():
        try:
            return original_done_current()
        finally:
            mutation_state["active"] = False

    monkeypatch.setattr(window.viewport, "makeCurrent", tracked_make_current)
    monkeypatch.setattr(window.viewport, "doneCurrent", tracked_done_current)
    monkeypatch.setattr(window.properties_panel, "set_snapshot", tracked_set_snapshot)

    original_create_scene = window.viewport.host.create_scene

    def counted_create_scene(scene_cls, system):
        scene = original_create_scene(scene_cls, system)
        scene._mutation_state = mutation_state
        return scene

    monkeypatch.setattr(window.viewport.host, "create_scene", counted_create_scene)

    document = _make_document(
        "field.sad",
        System.cubic((4, 4, 4)),
        [np.zeros((4, 4, 4, 1, 3), dtype=np.float32)],
    )

    window._activate_entry(document, document.entries[0])
    qapp.processEvents()
    property_calls.clear()

    window._on_property_edited("Vector", 0, "Arrow", "mesh_shape", "Shape")
    qapp.processEvents()

    assert property_calls == [False]
    window.close()


def test_qt_structural_property_edit_rebuilds_panel_once(qapp, monkeypatch):
    window = TvViewWindow()
    window.viewport._ctx = object()
    property_calls = []

    monkeypatch.setattr(
        "topovec.viewer.qt.main_window.SCENES",
        {"Layer": _QtSchemaScene},
    )
    monkeypatch.setattr(
        "topovec.viewer.session._load_builtin_scene_registry",
        lambda: {"Layer": _QtSchemaScene},
    )

    original_set_snapshot = window.properties_panel.set_snapshot

    def tracked_set_snapshot(snapshot, *, rebuild=True):
        property_calls.append(rebuild)
        return original_set_snapshot(snapshot, rebuild=rebuild)

    monkeypatch.setattr(window.properties_panel, "set_snapshot", tracked_set_snapshot)

    document = _make_document(
        "field.sad",
        System.cubic((4, 4, 4)),
        [np.zeros((4, 4, 4, 1, 3), dtype=np.float32)],
    )

    window._activate_entry(document, document.entries[0])
    qapp.processEvents()
    property_calls.clear()

    window._on_property_edited("Render", 0, "beta", "mode", "Mode")
    qapp.processEvents()

    assert property_calls == [True]
    window.close()


def test_cached_live_scene_is_restored_after_switching_to_different_system(
    qapp, monkeypatch
):
    window = TvViewWindow()
    window.viewport._ctx = object()
    create_calls = []

    _install_qt_scene_registry(monkeypatch)
    original_create_scene = window.viewport.host.create_scene

    def counted_create_scene(scene_cls, system):
        create_calls.append(getattr(scene_cls, "__name__", repr(scene_cls)))
        return original_create_scene(scene_cls, system)

    monkeypatch.setattr(window.viewport.host, "create_scene", counted_create_scene)

    document_a = _make_document(
        "field_a.sad",
        System.cubic((4, 4, 4)),
        [
            np.zeros((4, 4, 4, 1, 3), dtype=np.float32),
            np.ones((4, 4, 4, 1, 3), dtype=np.float32),
        ],
    )
    document_b = _make_document(
        "field_b.sad",
        System.cubic((5, 4, 4)),
        [np.full((5, 4, 4, 1, 3), 2.0, dtype=np.float32)],
    )

    window._activate_entry(document_a, document_a.entries[0])
    qapp.processEvents()

    session_a = window.session
    assert session_a is not None
    assert session_a.switch_scene("Ball") is True
    ball_scene = session_a.scene
    session_a.apply_property_change(
        "Render",
        0,
        True,
        expected_field="flag",
        expected_title="Flag",
    )
    session_a.camera.look_scale *= 1.75
    session_a.commit_camera_interaction()
    camera_a = camera_state_signature(session_a.camera_snapshot())

    window._activate_entry(document_b, document_b.entries[0])
    qapp.processEvents()

    session_b = window.session
    assert session_b is not None
    assert session_b is not session_a

    window._activate_entry(document_a, document_a.entries[1])
    qapp.processEvents()

    assert window.session is not session_a
    assert window.session is not session_b
    assert window.session.scene is ball_scene
    assert window.session.selected_scene == "Ball"
    assert camera_state_signature(window.session.camera_snapshot()) == camera_a
    props = window.session.properties_snapshot()["property_groups"][0]["properties"]
    assert props[0]["value"] is True
    assert create_calls == ["_QtLayerScene", "_QtBallScene", "_QtLayerScene"]
    window.close()


def test_different_system_activation_creates_new_session(qapp, monkeypatch):
    window = TvViewWindow()
    window.viewport._ctx = object()
    create_calls = []

    _install_qt_scene_registry(monkeypatch)
    original_create_scene = window.viewport.host.create_scene

    def counted_create_scene(scene_cls, system):
        create_calls.append(getattr(scene_cls, "__name__", repr(scene_cls)))
        return original_create_scene(scene_cls, system)

    monkeypatch.setattr(window.viewport.host, "create_scene", counted_create_scene)

    document_a = _make_document(
        "field_a.sad",
        System.cubic((4, 4, 4)),
        [np.zeros((4, 4, 4, 1, 3), dtype=np.float32)],
    )
    document_b = _make_document(
        "field_b.sad",
        System.cubic((5, 4, 4)),
        [np.ones((5, 4, 4, 1, 3), dtype=np.float32)],
    )

    window._activate_entry(document_a, document_a.entries[0])
    qapp.processEvents()
    session = window.session

    window._activate_entry(document_b, document_b.entries[0])
    qapp.processEvents()

    assert session is not None
    assert window.session is not session
    assert create_calls == ["_QtLayerScene", "_QtLayerScene"]
    window.close()


def test_cached_scene_restore_failure_keeps_previous_active_session(qapp, monkeypatch):
    window = TvViewWindow()
    window.viewport._ctx = object()

    _install_qt_scene_registry(monkeypatch)
    document_a = _make_document(
        "field_a.sad",
        System.cubic((4, 4, 4)),
        [
            np.zeros((4, 4, 4, 1, 3), dtype=np.float32),
            np.ones((4, 4, 4, 1, 3), dtype=np.float32),
        ],
    )
    document_b = _make_document(
        "field_b.sad",
        System.cubic((5, 4, 4)),
        [np.full((5, 4, 4, 1, 3), 2.0, dtype=np.float32)],
    )

    window._activate_entry(document_a, document_a.entries[0])
    qapp.processEvents()

    session_a = window.session
    assert session_a is not None

    window._activate_entry(document_b, document_b.entries[0])
    qapp.processEvents()

    session_b = window.session
    assert session_b is not None
    camera_before = camera_state_signature(session_b.camera_snapshot())
    monkeypatch.setattr(
        ViewerSession,
        "restore_residency_snapshot",
        lambda self, _snapshot: (_ for _ in ()).throw(RuntimeError("cache boom")),
    )

    window._activate_entry(document_a, document_a.entries[1])
    qapp.processEvents()

    assert window.session is session_b
    assert window._current_document is document_b
    assert window._current_entry is document_b.entries[0]
    assert camera_state_signature(window.session.camera_snapshot()) == camera_before
    assert "Failed to activate" in window.statusBar().currentMessage()
    assert "cache boom" in window.log_panel.text()
    window.close()


def test_closing_current_source_keeps_cached_scene_for_same_system(qapp, monkeypatch):
    window = TvViewWindow()
    window.viewport._ctx = object()
    create_calls = []

    _install_qt_scene_registry(monkeypatch)
    original_create_scene = window.viewport.host.create_scene

    def counted_create_scene(scene_cls, system):
        create_calls.append(getattr(scene_cls, "__name__", repr(scene_cls)))
        return original_create_scene(scene_cls, system)

    monkeypatch.setattr(window.viewport.host, "create_scene", counted_create_scene)

    document_a = _make_document(
        "field_a.sad",
        System.cubic((4, 4, 4)),
        [np.zeros((4, 4, 4, 1, 3), dtype=np.float32)],
    )
    document_b = _make_document(
        "field_b.sad",
        System.cubic((4, 4, 4)),
        [np.ones((4, 4, 4, 1, 3), dtype=np.float32)],
    )

    documents = {
        "/tmp/field_a.sad": document_a,
        "/tmp/field_b.sad": document_b,
    }
    monkeypatch.setattr(
        "topovec.viewer.qt.main_window.load_viewer_document",
        lambda raw_path: documents[str(raw_path)],
    )

    window.load_paths(["/tmp/field_a.sad", "/tmp/field_b.sad"])
    qapp.processEvents()

    session = window.session
    assert session is not None
    active_scene = session.scene
    session.camera.look_scale *= 1.5
    session.commit_camera_interaction()
    camera_before = camera_state_signature(session.camera_snapshot())

    first_state_item = window.sources_tree.topLevelItem(0).child(0)
    window.sources_tree.setCurrentItem(first_state_item)
    qapp.processEvents()
    window.close_selected_source()
    qapp.processEvents()

    assert window.session is not session
    assert window.session.scene is active_scene
    assert window._current_document is document_b
    assert window._current_entry is document_b.entries[0]
    assert camera_state_signature(window.session.camera_snapshot()) == camera_before
    assert create_calls == ["_QtLayerScene"]
    window.close()


def test_viewport_shutdown_releases_context(qapp, monkeypatch):
    viewport = QtViewportWindow()
    calls = []

    class DummyContext:
        def gc(self):
            calls.append("gc")

        def release(self):
            calls.append("release")

    monkeypatch.setattr(viewport, "makeCurrent", lambda: calls.append("makeCurrent"))
    monkeypatch.setattr(viewport, "doneCurrent", lambda: calls.append("doneCurrent"))
    viewport._ctx = DummyContext()
    viewport._session = object()
    viewport._scene = object()

    viewport.shutdown()

    assert calls == [
        "makeCurrent",
        "gc",
        "release",
        "doneCurrent",
    ]
    assert viewport._ctx is None
    assert viewport._scene is None
    assert viewport._session is None



import logging
import sys

log = logging.getLogger("topovec")


def configLog(level=logging.DEBUG):
    global log
    log.setLevel(level)
    log.propagate = False

    for handler in log.handlers:
        if getattr(handler, "_topovec_console_handler", False):
            handler.setLevel(level)
            return log

    try:
        from rich.console import Console
        from rich.logging import RichHandler

        handler = RichHandler(
            rich_tracebacks=True,
            console=Console(stderr=True),
            show_path=False,
        )
    except Exception:
        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(
            logging.Formatter(
                fmt="%(asctime)s %(levelname)s %(name)s: %(message)s",
                datefmt="%H:%M:%S",
            )
        )
    handler.setLevel(level)
    handler._topovec_console_handler = True
    log.addHandler(handler)
    return log



from types import SimpleNamespace

import moderngl as mgl
import numpy as np

import topovec.mgl as topovec_mgl
from topovec.core import System
from topovec.mgl.axes import AxesRender
from topovec.mgl.charge import ChargeRender
from topovec.mgl.scene import Scene
from topovec.mgl.scene import ChargeScene, EmptyScene, FlowScene, LayerAndSurfaceScene, VectorScene
from topovec.mgl.target import RenderTarget


class DummyScene(Scene):
    def upload(self, state=None):
        return None

    def render(self):
        return None


def test_scene_clear_background_uses_current_framebuffer():
    clear_calls = []

    class DummyFramebuffer:
        def clear(self, *args, **kwargs):
            clear_calls.append((args, kwargs))

    ctx = SimpleNamespace(fbo=DummyFramebuffer())
    system = SimpleNamespace()

    scene = DummyScene.__new__(DummyScene)
    scene._ctx = ctx
    scene._system = system
    scene._background = (0.25, 0.5, 0.75)

    scene.clear_background()
    scene.clear_current_framebuffer(0.1, 0.2, 0.3, 0.4)

    assert clear_calls == [
        ((0.25, 0.5, 0.75, 1.0), {}),
        ((0.1, 0.2, 0.3, 0.4), {}),
    ]


def test_scene_render_compatibility_wrapper_routes_target_based_scene():
    class DummyFramebuffer:
        def __init__(self):
            self.size = (640, 320)
            self.use_calls = 0

        def use(self):
            self.use_calls += 1

    class TargetScene(Scene):
        def upload(self, state=None):
            return None

        def render_to(self, target):
            self.last_target = target

    framebuffer = DummyFramebuffer()
    ctx = SimpleNamespace(fbo=framebuffer, viewport=None)
    scene = TargetScene.__new__(TargetScene)
    scene._ctx = ctx
    scene._system = SimpleNamespace()
    scene._background = (1.0, 1.0, 1.0)

    scene.render(aspectratio=2.5)

    assert isinstance(scene.last_target, RenderTarget)
    assert scene.last_target.aspectratio == 2.5
    assert scene.last_target.size == (640, 320)


def test_scene_render_to_falls_back_to_legacy_render_override():
    class LegacyScene(Scene):
        def upload(self, state=None):
            return None

        def render(self, aspectratio=None):
            self.last_aspectratio = aspectratio

    scene = LegacyScene.__new__(LegacyScene)
    scene._ctx = SimpleNamespace(fbo=SimpleNamespace(size=(100, 50)))
    scene._system = SimpleNamespace()
    target = RenderTarget(
        ctx=SimpleNamespace(viewport=None),
        fbo=SimpleNamespace(size=(100, 50), color_mask=(True, True, True, True), depth_mask=True),
        size=(100, 50),
        aspectratio=3.0,
    )

    Scene.render_to(scene, target)

    assert scene.last_aspectratio == 3.0


def test_axes_render_release_does_not_invalidate_shared_context():
    system = System.cubic((4, 4, 4))
    state = np.zeros((4, 4, 4, 3), dtype=np.float32)
    state[..., 2] = 1.0

    image_container = topovec_mgl.render_prepare(
        scenecls="Layer",
        system=system,
        imgsize=32,
    )
    image_container.upload(state)

    assert image_container.save().size == image_container.size

    image_container.scene.axes_render.release()

    fresh_axes = AxesRender(mglctx=image_container._ctx)

    assert fresh_axes._prog is not None


def test_scene_release_keeps_shared_context_usable_for_new_scene():
    system = System.cubic((4, 4, 4))
    state = np.zeros((4, 4, 4, 3), dtype=np.float32)
    state[..., 2] = 1.0

    image_container = topovec_mgl.render_prepare(
        scenecls="Layer",
        system=system,
        imgsize=32,
    )
    image_container.upload(state)

    assert image_container.save().size == image_container.size

    image_container.scene.release()

    replacement = topovec_mgl.SCENES["Level surface"](
        mglctx=image_container._ctx,
        system=system,
    )
    image_container.attach_scene(replacement)
    image_container.upload(state)

    assert image_container.save().size == image_container.size


class DummyAxes:
    def list_properties(self):
        return []

    def render(self, **kwargs):
        return None


class DummyCamera:
    def get_view_matrix(self):
        return np.eye(4, dtype=np.float32)

    def get_projection_matrix(self, aspectratio=None):
        return np.eye(4, dtype=np.float32)

    def camera_position(self):
        return (0.0, 0.0, 1.0)


class DummyTarget:
    def __init__(self):
        self.aspectratio = 1.5
        self.clear_calls = []
        self.bind_calls = 0

    def bind(self):
        self.bind_calls += 1

    def clear_color(self, *rgba):
        if len(rgba) == 1 and not isinstance(rgba[0], (int, float)):
            rgba = tuple(rgba[0])
        self.clear_calls.append(tuple(float(value) for value in rgba))


class DummyRender:
    MESH_NAMES = {"Cone": object(), "Arrow": object()}
    ANGLENAMES = ["polar angle", "azimuthal angle"]

    def __init__(self, **attrs):
        self.render_calls = []
        self.color_shift = 0.81
        self.lighting = True
        self.negate = False
        self.width = 0.5
        self.max_size = 1.0
        self.neglect_z = False
        self.level = 0.0
        self.angle_name = self.ANGLENAMES[0]
        self.mesh_shape = "Cone"
        self.subtract_blend = True
        for name, value in attrs.items():
            setattr(self, name, value)

    def render(self, **kwargs):
        self.render_calls.append(dict(kwargs))


class DummyContext:
    def __init__(self):
        self.enable_only_calls = []
        self.blend_func = None
        self.blend_equation = None
        self.line_width = 0.0

    def enable_only(self, flag):
        self.enable_only_calls.append(flag)


class DummyUniform:
    def __init__(self):
        self.value = None


class DummyVao:
    def __init__(self):
        self.calls = []

    def render(self, **kwargs):
        self.calls.append(dict(kwargs))


def _empty_scene_stub(scene_cls):
    scene = scene_cls.__new__(scene_cls)
    scene._background = (1.0, 1.0, 1.0)
    scene._white_background = True
    scene._ctx = DummyContext()
    scene._axes = DummyAxes()
    scene.show_axes = False
    scene._camera = DummyCamera()
    return scene


def test_empty_scene_style_group_exposes_white_background():
    scene = _empty_scene_stub(EmptyScene)

    groups = EmptyScene.list_properties(scene)

    assert list(groups)[:2] == ["Style", "Axes"]
    style_prop = groups["Style"][0]
    assert style_prop.field == "white_background"
    assert style_prop.title == "White background"
    assert style_prop.param_key == "style.white_background"

    scene.white_background = False
    assert scene._background == (0.0, 0.0, 0.0)
    scene.white_background = True
    assert scene._background == (1.0, 1.0, 1.0)


def test_vector_scene_render_uses_common_white_background_toggle():
    scene = _empty_scene_stub(VectorScene)
    scene._render = DummyRender()
    target = DummyTarget()

    VectorScene.render_to(scene, target)
    scene.white_background = False
    VectorScene.render_to(scene, target)

    assert target.clear_calls == [
        (1.0, 1.0, 1.0),
        (0.0, 0.0, 0.0),
    ]
    assert len(scene._render.render_calls) == 2


def test_flow_scene_render_uses_common_white_background_toggle():
    scene = _empty_scene_stub(FlowScene)
    scene._render = DummyRender()
    target = DummyTarget()

    scene.white_background = False
    FlowScene.render_to(scene, target)
    scene.white_background = True
    FlowScene.render_to(scene, target)

    assert target.clear_calls == [
        (0.0, 0.0, 0.0),
        (1.0, 1.0, 1.0),
    ]
    assert len(scene._render.render_calls) == 2


def test_layer_and_surface_scene_uses_style_group_and_common_background():
    scene = _empty_scene_stub(LayerAndSurfaceScene)
    scene._render = DummyRender()
    scene._level_render = DummyRender()
    scene.layers_count = 4
    target = DummyTarget()

    groups = LayerAndSurfaceScene.list_properties(scene)
    LayerAndSurfaceScene.render_to(scene, target)
    scene.white_background = False
    LayerAndSurfaceScene.render_to(scene, target)

    assert "Style" in groups
    assert "Shared" not in groups
    assert groups["Style"][0].field == "white_background"
    assert target.clear_calls == [
        (1.0, 1.0, 1.0),
        (0.0, 0.0, 0.0),
    ]
    assert len(scene._render.render_calls) == 2
    assert len(scene._level_render.render_calls) == 2


def test_charge_scene_background_follows_white_background_only():
    scene = _empty_scene_stub(ChargeScene)
    scene._render = DummyRender(subtract_blend=True)
    target = DummyTarget()

    scene.white_background = False
    ChargeScene.render_to(scene, target)
    scene.white_background = True
    ChargeScene.render_to(scene, target)

    assert target.clear_calls == [
        (0.0, 0.0, 0.0),
        (1.0, 1.0, 1.0),
    ]


def test_charge_render_uses_subtract_blend_property_for_blend_equation():
    render = ChargeRender.__new__(ChargeRender)
    render._ctx = DummyContext()
    render._vao = DummyVao()
    render.u_view = DummyUniform()
    render.u_projection = DummyUniform()
    render.u_brightness = DummyUniform()
    render.u_subtract_blend = DummyUniform()
    render.brightness = 2.0
    modelview = np.eye(4, dtype=np.float32)
    projection = np.eye(4, dtype=np.float32)

    render.subtract_blend = True
    ChargeRender.render(
        render,
        modelview=modelview,
        projection=projection,
        camera_position=(0.0, 0.0, 1.0),
    )
    min_equation = render._ctx.blend_equation

    render.subtract_blend = False
    ChargeRender.render(
        render,
        modelview=modelview,
        projection=projection,
        camera_position=(0.0, 0.0, 1.0),
    )

    props = ChargeRender.list_properties(render)

    assert min_equation == mgl.MIN
    assert render._ctx.blend_equation == mgl.MAX
    assert props[0].field == "subtract_blend"
    assert props[0].title == "Subtract blend"



"""Backend-neutral viewer session and host abstractions.

This module owns the reusable controller logic shared by notebook and desktop
frontends. It manages scene switching, camera state, property snapshots, and
Python snippet generation without depending on marimo, Qt, or browser
transport.
"""

from __future__ import annotations

import copy
from contextlib import nullcontext
from dataclasses import dataclass
from typing import Any, Callable, Protocol, runtime_checkable

import numpy as np
from PIL import Image

from ..core import ListProperty, NumericProperty, Property, System, UpdateMask
from ..mgl.camera import Camera

PROJECTION_LABELS = ("Perspective", "Orthographic")
DEFAULT_NUMERIC_STEP = 0.001


@dataclass(slots=True)
class ViewerFrame:
    """One rendered frame plus viewer metadata."""

    image: Image.Image
    status_text: str
    projection_mode: str
    camera_state: dict[str, Any]
    width: int
    height: int


@dataclass(slots=True)
class ViewerSessionResidencySnapshot:
    """Lightweight state retained across live-scene eviction or session rebuild."""

    shared_state: dict[str, Any]
    selected_scene: str | None
    default_property_values_by_scene: dict[str, dict[tuple[str, str, str], Any]]


@runtime_checkable
class ViewerHost(Protocol):
    """Minimal backend interface required by :class:`ViewerSession`."""

    @property
    def scene(self) -> Any: ...

    @property
    def size(self) -> tuple[int, int]: ...

    def create_scene(self, scene_cls: Any, system: Any) -> Any: ...

    def attach_scene(self, scene: Any) -> Any: ...

    def upload(self, state: np.ndarray | None) -> Any: ...

    def scene_mutation_scope(self, reason: str): ...

    def render_image(
        self,
        *,
        size: tuple[int, int] | None = None,
    ) -> Image.Image: ...


class ImageContainerHost:
    """Adapt :class:`topovec.mgl.image.ImageContainer` to :class:`ViewerHost`."""

    def __init__(self, image_container: Any):
        self._image_container = image_container

    @property
    def raw(self) -> Any:
        return self._image_container

    @property
    def scene(self) -> Any:
        return self._image_container.scene

    @property
    def size(self) -> tuple[int, int]:
        size = getattr(
            self._image_container,
            "size",
            getattr(self._image_container, "_size", (512, 512)),
        )
        return int(size[0]), int(size[1])

    def create_scene(self, scene_cls: Any, system: Any) -> Any:
        context = getattr(self._image_container, "_ctx", None)
        return scene_cls(mglctx=context, system=system)

    def attach_scene(self, scene: Any) -> Any:
        self._image_container.attach_scene(scene)
        return self

    def upload(self, state: np.ndarray | None) -> Any:
        self._image_container.upload(state)
        return self

    def scene_mutation_scope(self, reason: str):
        return nullcontext()

    def render_image(
        self,
        *,
        size: tuple[int, int] | None = None,
    ) -> Image.Image:
        requested = None if size is None else (int(size[0]), int(size[1]))
        if requested is not None and requested != self.size:
            raise NotImplementedError(
                "ImageContainerHost only renders at the container framebuffer size."
            )
        return self._image_container.save()


def adapt_viewer_host(host: Any) -> ViewerHost:
    """Normalize a raw image container or host-like object."""
    if isinstance(host, ViewerHost):
        return host
    return ImageContainerHost(host)


def serialize_camera(camera: Camera) -> dict[str, Any]:
    """Convert a live camera into a JSON-friendly backend-neutral snapshot."""
    return {
        "look_at": [float(v) for v in np.asarray(camera.look_at, dtype=float)],
        "look_direction": [
            float(v) for v in np.asarray(camera.look_direction, dtype=float)
        ],
        "look_up": [float(v) for v in np.asarray(camera.look_up, dtype=float)],
        "look_scale": float(camera.look_scale),
        "perspective": bool(camera.perspective_is_on),
    }


def restore_camera(camera: Camera, state: dict[str, Any]) -> Camera:
    """Apply a serialized snapshot back to an existing camera instance."""
    camera.look_at = np.asarray(state["look_at"], dtype=np.float32)
    camera.look_direction = np.asarray(state["look_direction"], dtype=np.float32)
    camera.look_up = np.asarray(state["look_up"], dtype=np.float32)
    camera.look_scale = float(state["look_scale"])
    camera.perspective_is_on = bool(state["perspective"])
    return camera.normalize()


def _normalize_camera_state(camera_state: dict[str, Any]) -> dict[str, Any]:
    camera = Camera(perspective=bool(camera_state["perspective"]))
    return serialize_camera(restore_camera(camera, camera_state))


def camera_state_signature(camera_state: dict[str, Any]) -> tuple[Any, ...]:
    """Create a rounded signature used to compare camera states."""
    state = _normalize_camera_state(camera_state)

    def rounded(values: list[float]) -> tuple[float, ...]:
        return tuple(round(float(v), 6) for v in values)

    return (
        rounded(state["look_at"]),
        rounded(state["look_direction"]),
        rounded(state["look_up"]),
        round(float(state["look_scale"]), 6),
        bool(state["perspective"]),
    )


def _format_python_float(value: float) -> str:
    return format(float(value), ".6g")


def _coerce_snapshot_value(value: Any) -> Any:
    if isinstance(value, np.generic):
        return value.item()
    if isinstance(value, np.ndarray):
        return value.tolist()
    if isinstance(value, tuple):
        return [_coerce_snapshot_value(v) for v in value]
    if isinstance(value, list):
        return [_coerce_snapshot_value(v) for v in value]
    return value


def _format_python_value(value: Any) -> str:
    value = _coerce_snapshot_value(value)
    if isinstance(value, bool):
        return "True" if value else "False"
    if isinstance(value, (int, np.integer)) and not isinstance(value, bool):
        return str(int(value))
    if isinstance(value, (float, np.floating)):
        return _format_python_float(float(value))
    if isinstance(value, str):
        return repr(value)
    if isinstance(value, list):
        return "[" + ", ".join(_format_python_value(item) for item in value) + "]"
    if isinstance(value, tuple):
        inner = ", ".join(_format_python_value(item) for item in value)
        if len(value) == 1:
            inner += ","
        return f"({inner})"
    if value is None:
        return "None"
    return repr(value)


def _projection_label(perspective: bool) -> str:
    return PROJECTION_LABELS[0] if perspective else PROJECTION_LABELS[1]


def format_camera_summary(camera: Camera, *, topovec_name: str = "tv") -> str:
    """Render the current camera as executable Python code."""
    theta, phi, alpha = camera.get_orientation_angles()
    look_at = ", ".join(
        _format_python_value(v)
        for v in np.asarray(camera.look_at, dtype=float)
    )
    return (
        f"{topovec_name}.mgl.Camera.from_angles("
        f"theta={_format_python_value(theta)}, "
        f"phi={_format_python_value(phi)}, "
        f"alpha={_format_python_value(alpha)}, "
        f"scale={_format_python_value(camera.look_scale)}, "
        f"perspective={camera.perspective_is_on}, "
        f"look_at=[{look_at}])"
    )


def apply_camera_command(
    camera: Camera,
    *,
    event_type: str,
    dx: float = 0.0,
    dy: float = 0.0,
    wheel: float = 0.0,
    rotate_rate: float = 2 * np.pi,
    roll_rate: float = 0.02 * np.pi,
    pan_rate: float = 2.0,
    zoom_rate: float = 0.12,
) -> Camera:
    """Apply one normalized interaction event to a camera."""
    dx = float(dx)
    dy = float(dy)
    wheel = float(wheel)

    if event_type == "rotate":
        camera.rotate_field_of_view(x=-dx, y=dy, rate=rotate_rate)
    elif event_type == "pan":
        camera.move_field_of_view(x=-dx, y=dy, rate=pan_rate)
    elif event_type == "roll":
        camera.rotate_around(z=-1.0, rate=dx * roll_rate)
    elif event_type == "zoom":
        camera.look_scale *= float(np.exp(np.clip(wheel, -8.0, 8.0) * zoom_rate))
    elif event_type in {"", "noop"}:
        return camera
    else:
        raise ValueError(f"Unsupported camera event: {event_type!r}")
    return camera


def _copy_shared_value(value: Any) -> Any:
    return copy.deepcopy(_coerce_snapshot_value(value))


def _shared_values_equal(left: Any, right: Any) -> bool:
    if isinstance(left, np.ndarray) or isinstance(right, np.ndarray):
        return np.array_equal(np.asarray(left), np.asarray(right))
    try:
        return bool(left == right)
    except Exception:
        return False


def _coerce_director_field_state(state: Any) -> np.ndarray:
    array = np.asarray(state)
    if array.ndim < 4:
        raise ValueError(
            "`inspect` expects a director field with at least three spatial "
            f"dimensions and a trailing vector axis, got shape {array.shape!r}."
        )
    if array.shape[-1] != 3:
        raise ValueError(
            "`inspect` expects the last state axis to have length 3, "
            f"got shape {array.shape!r}."
        )
    return array


def infer_cubic_system_from_state(state: np.ndarray) -> System:
    """Infer a trivial cubic system from a vector field shape."""
    return System.cubic(size=tuple(int(v) for v in state.shape[:3]))


def _load_builtin_scene_registry() -> dict[str, Any]:
    try:
        from ..mgl import SCENES
    except Exception:
        return {}
    return dict(SCENES)


def _scene_class_supports_system(scene_cls: Any, system: Any) -> bool:
    supports_system = getattr(scene_cls, "supports_system", None)
    if not callable(supports_system):
        return True
    return bool(supports_system(system))


def _resolve_builtin_scene_label(
    scene: Any,
    registry: dict[str, Any],
) -> str | None:
    scene_cls = getattr(scene, "__class__", None)
    for raw_label, candidate in registry.items():
        if scene_cls is candidate:
            return str(raw_label)
    return None


def _property_snapshot(prop: Property, index: int) -> dict[str, Any]:
    spec = dict(prop.spec())
    snapshot = {
        "index": int(index),
        "kind": str(spec["kind"]),
        "field": str(spec["field"]),
        "title": str(spec["title"]),
        "value": _coerce_snapshot_value(spec["value"]),
        "invalidates": bool(spec["invalidates"]),
        "param_key": spec.get("param_key"),
    }
    if isinstance(prop, NumericProperty):
        snapshot.update(
            {
                "min": float(spec["min"]),
                "max": float(spec["max"]),
                "count": int(spec["count"]),
            }
        )
    elif isinstance(prop, ListProperty):
        snapshot.update(
            {
                "values": [_coerce_snapshot_value(v) for v in spec["values"]],
                "selected_index": int(spec["selected_index"]),
            }
        )
    return snapshot


def _numeric_step(prop_snapshot: dict[str, Any]) -> float | None:
    count = int(prop_snapshot.get("count", 0))
    minimum = float(prop_snapshot.get("min", 0.0))
    maximum = float(prop_snapshot.get("max", 0.0))
    if count <= 1:
        return None
    delta = maximum - minimum
    if not np.isfinite(delta) or delta <= 0:
        return None
    step = delta / float(count - 1)
    if not np.isfinite(step) or step <= 0:
        return None
    return float(step)


def _default_numeric_step() -> float:
    return DEFAULT_NUMERIC_STEP


def _effective_numeric_step(prop_snapshot: dict[str, Any]) -> float:
    step = _numeric_step(prop_snapshot)
    if step is not None:
        return step
    return _default_numeric_step()


def _normalized_numeric_ui_step(
    step: float | None,
    *,
    significant_digits: int = 2,
) -> float | None:
    """Round a raw numeric step to a short terminating decimal for UI widgets."""
    if step is None:
        return None
    step = float(abs(step))
    if not np.isfinite(step) or step <= 0:
        return None

    precision = max(
        int(significant_digits - np.floor(np.log10(step)) - 1),
        0,
    )
    normalized = round(step, precision)
    while normalized == 0.0:
        precision += 1
        normalized = round(step, precision)
    return float(normalized)


def _numeric_bounds(
    prop_snapshot: dict[str, Any],
) -> tuple[float, float, float, bool]:
    """Normalize numeric metadata and report whether the current value fits it."""
    minimum = float(prop_snapshot.get("min", 0.0))
    maximum = float(prop_snapshot.get("max", 0.0))
    value = float(prop_snapshot.get("value", 0.0))

    if np.isfinite(minimum) and np.isfinite(maximum) and minimum > maximum:
        minimum, maximum = maximum, minimum

    value_in_bounds = (
        np.isfinite(minimum)
        and np.isfinite(maximum)
        and minimum <= value <= maximum
    )
    return minimum, maximum, value, value_in_bounds


def _numeric_step_decimals(step: float | None) -> int | None:
    if step is None:
        return None
    step = float(abs(step))
    if not np.isfinite(step) or step <= 0:
        return None
    text = f"{step:.12f}".rstrip("0").rstrip(".")
    if "." not in text:
        return 0
    return len(text.split(".", 1)[1])


def _quantize_numeric_value(value: float, prop_snapshot: dict[str, Any]) -> float:
    """Snap numeric input to either count-based or fallback resolution."""
    step = _effective_numeric_step(prop_snapshot)
    minimum, maximum, _, value_in_bounds = _numeric_bounds(prop_snapshot)
    snapped = float(value)

    if value_in_bounds:
        snapped = minimum + round((snapped - minimum) / step) * step
        snapped = float(np.clip(snapped, minimum, maximum))
    else:
        snapped = round(snapped / step) * step

    decimals = _numeric_step_decimals(step)
    if decimals is not None:
        snapped = round(snapped, decimals)

    return float(snapped)


def _clamp_numeric_value(value: float, prop_snapshot: dict[str, Any]) -> float:
    """Clamp numeric input to declared bounds without snapping to the step grid."""
    minimum, maximum, _current, _value_in_bounds = _numeric_bounds(prop_snapshot)
    clamped = float(value)
    if np.isfinite(minimum):
        clamped = max(float(minimum), clamped)
    if np.isfinite(maximum):
        clamped = min(float(maximum), clamped)
    return float(clamped)


def _decorate_property_snapshot(prop_snapshot: dict[str, Any]) -> dict[str, Any]:
    snapshot = dict(prop_snapshot)
    kind = str(snapshot.get("kind", ""))

    if kind == "numeric":
        minimum, maximum, value, value_in_bounds = _numeric_bounds(snapshot)
        step = _normalized_numeric_ui_step(_effective_numeric_step(snapshot))
        snapshot["value"] = float(value)
        snapshot["step"] = step
        snapshot["value_in_bounds"] = bool(value_in_bounds)
        snapshot["input_kind"] = "number"
    elif kind == "boolean":
        snapshot["input_kind"] = "switch"
    elif kind == "list":
        snapshot["input_kind"] = "dropdown"
    else:
        snapshot["input_kind"] = "display"

    snapshot["row_kind"] = "inline"

    return snapshot


class ViewerSession:
    """Python-side source of truth for one live render session."""

    def __init__(
        self,
        host: Any,
        state: np.ndarray | None = None,
        *,
        controls_text: str = "",
        rotate_rate: float = 2 * np.pi,
        roll_rate: float = 0.02 * np.pi,
        pan_rate: float = 2.0,
        zoom_rate: float = 0.12,
    ):
        self._host = adapt_viewer_host(host)
        self._scene = self._host.scene
        self._camera = self._scene.camera
        self.controls_text = controls_text
        self.rotate_rate = float(rotate_rate)
        self.roll_rate = float(roll_rate)
        self.pan_rate = float(pan_rate)
        self.zoom_rate = float(zoom_rate)
        self._revision = 0
        self._state = state
        self._committed_camera_state = serialize_camera(self._camera)
        self._shared_state: dict[str, Any] = {}
        self._scene_cache: dict[str, Any] = {}
        self._default_property_values_by_scene: dict[
            str, dict[tuple[str, str, str], Any]
        ] = {}
        self._selected_scene: str | None = None
        self._available_scenes: list[str] = []
        self._property_groups: list[dict[str, Any]] = []
        self._property_group_index: dict[str, int] = {}
        self._properties_by_group: dict[str, list[Property]] = {}

        if state is not None:
            self._host.upload(state)

        self._refresh_scene_metadata()
        self._cache_current_scene()
        self._capture_current_scene_default_baseline()
        self._sync_active_scene_from_shared_state()
        self._restore_or_register_current_camera_slot()
        self._committed_camera_state = self._current_camera_state()
        self._refresh_property_cache()

    @property
    def host(self) -> ViewerHost:
        return self._host

    @property
    def image_container(self) -> ViewerHost:
        """Backward-compatible alias used by the marimo adapter."""
        return self._host

    @property
    def scene(self) -> Any:
        return self._scene

    @property
    def camera(self) -> Camera:
        return self._camera

    @property
    def revision(self) -> int:
        return self._revision

    @property
    def can_undo(self) -> bool:
        return False

    @property
    def can_redo(self) -> bool:
        return False

    @property
    def selected_scene(self) -> str | None:
        return self._selected_scene

    @property
    def available_scenes(self) -> list[str]:
        return list(self._available_scenes)

    def _scene_system(self) -> Any:
        return getattr(self._scene, "system", None)

    def _scene_camera_slot(self) -> str:
        slot = getattr(
            self._scene,
            "camera_slot_key",
            getattr(self._scene, "camera_profile", "general"),
        )
        return str(slot or "general")

    def _camera_key(self, slot: str) -> str:
        return f"camera:{slot}"

    def _home_camera_key(self, slot: str) -> str:
        return f"camera-home:{slot}"

    def _current_camera_state(self) -> dict[str, Any]:
        return serialize_camera(self._camera)

    def _store_camera_state(
        self,
        slot: str,
        camera_state: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        state = _normalize_camera_state(
            self._current_camera_state() if camera_state is None else camera_state
        )
        self._shared_state[self._camera_key(str(slot))] = dict(state)
        return dict(state)

    def _ensure_home_state(
        self,
        slot: str,
        camera_state: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        slot = str(slot)
        key = self._home_camera_key(slot)
        if key not in self._shared_state:
            state = _normalize_camera_state(
                self._current_camera_state() if camera_state is None else camera_state
            )
            self._shared_state[key] = dict(state)
        return dict(self._shared_state[key])

    def _register_current_camera_slot(self) -> dict[str, Any]:
        slot = self._scene_camera_slot()
        self._activate_current_scene_camera_slot()
        state = self._store_camera_state(slot)
        self._ensure_home_state(slot, state)
        return state

    def _restore_or_register_current_camera_slot(self) -> bool:
        slot = self._scene_camera_slot()
        self._activate_current_scene_camera_slot()
        cached = self._shared_state.get(self._camera_key(slot))
        if cached is None:
            self._register_current_camera_slot()
            return False
        restore_camera(self._camera, cached)
        self._ensure_home_state(slot, cached)
        return True

    def _set_committed_camera_state(
        self,
        camera_state: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        state = _normalize_camera_state(
            self._current_camera_state() if camera_state is None else camera_state
        )
        self._committed_camera_state = dict(state)
        slot = self._scene_camera_slot()
        self._store_camera_state(slot, state)
        self._ensure_home_state(slot, state)
        return dict(state)

    def _cache_current_scene(self) -> None:
        if self._selected_scene is not None:
            self._scene_cache[self._selected_scene] = self._scene

    def _property_baseline_key(
        self,
        group_id: str,
        prop_snapshot: dict[str, Any],
    ) -> tuple[str, str, str]:
        return (
            str(group_id),
            str(prop_snapshot.get("field", "")),
            str(prop_snapshot.get("title", "")),
        )

    def _builtin_scene_registry(self) -> dict[str, Any]:
        return _load_builtin_scene_registry()

    def _refresh_scene_metadata(self) -> None:
        registry = self._builtin_scene_registry()
        self._selected_scene = _resolve_builtin_scene_label(self._scene, registry)
        self._cache_current_scene()

        system = self._scene_system()
        if system is None:
            self._available_scenes = []
            return

        self._available_scenes = [
            str(raw_label)
            for raw_label, scene_cls in registry.items()
            if _scene_class_supports_system(scene_cls, system)
        ]

    def _create_scene(self, scene_cls: Any) -> Any:
        system = self._scene_system()
        if system is None:
            raise RuntimeError("Scene switching requires `scene.system`.")
        return self._host.create_scene(scene_cls, system)

    def _attach_scene(self, scene: Any) -> None:
        self._host.attach_scene(scene)
        self._scene = self._host.scene
        self._camera = self._scene.camera

    def _activate_current_scene_camera_slot(self) -> None:
        activate = getattr(self._scene, "activate_camera_slot", None)
        if callable(activate):
            activate()
        self._camera = self._scene.camera

    def _register_property_shared_value(self, prop: Property) -> None:
        if prop.param_key is None:
            return
        self._shared_state[prop.param_key] = _copy_shared_value(prop.get())

    def _sync_active_scene_from_shared_state(self) -> UpdateMask:
        aggregate = UpdateMask.NONE
        with self._host.scene_mutation_scope("sync_active_scene_from_shared_state"):
            for _ in range(4):
                props_by_group = self._scene.list_properties()
                iteration = UpdateMask.NONE
                changed = False
                previous_slot = self._scene_camera_slot()
                previous_camera = self._current_camera_state()
                for props in props_by_group.values():
                    for prop in props:
                        key = prop.param_key
                        if key is None:
                            continue
                        if key not in self._shared_state:
                            self._register_property_shared_value(prop)
                            continue
                        target = _copy_shared_value(self._shared_state[key])
                        if _shared_values_equal(prop.get(), target):
                            continue
                        iteration |= UpdateMask(prop.let(target))
                        changed = True
                if iteration & UpdateMask.CAMERA_SLOT_CHANGED:
                    self._store_camera_state(previous_slot, previous_camera)
                    self._restore_or_register_current_camera_slot()
                aggregate |= iteration
                if not changed or not (iteration & UpdateMask.REBUILD_PROPERTIES):
                    break
        return aggregate

    def has_pending_camera_delta(self) -> bool:
        return camera_state_signature(
            self._current_camera_state()
        ) != camera_state_signature(self._committed_camera_state)

    def _increment_revision(self) -> None:
        self._revision += 1

    def set_home(self) -> None:
        slot = self._scene_camera_slot()
        state = self._current_camera_state()
        self._shared_state[self._home_camera_key(slot)] = _normalize_camera_state(state)
        self._store_camera_state(slot, state)

    def frame_snapshot(self) -> ViewerFrame:
        """Render the current frame into a transport-neutral snapshot."""
        image = self._host.render_image()
        size = self._host.size
        return ViewerFrame(
            image=image,
            status_text=format_camera_summary(self._camera),
            projection_mode=_projection_label(self._camera.perspective_is_on),
            camera_state=self._current_camera_state(),
            width=int(size[0]),
            height=int(size[1]),
        )

    def frame_payload(
        self,
        *,
        image_encoder: Callable[[Image.Image], Any] | None = None,
    ) -> dict[str, Any]:
        """Render the current frame and package it for frontend transport."""
        frame = self.frame_snapshot()
        payload = {
            "status_text": frame.status_text,
            "projection_mode": frame.projection_mode,
            "camera_state": dict(frame.camera_state),
            "width": frame.width,
            "height": frame.height,
        }
        if image_encoder is None:
            payload["image"] = frame.image
        else:
            payload["image_data"] = image_encoder(frame.image)
        return payload

    def export_image(
        self,
        *,
        size: tuple[int, int] | None = None,
    ) -> Image.Image:
        """Render the current scene to a PIL image, optionally at a custom size."""
        return self._host.render_image(size=size)

    def sync_viewer(
        self,
        viewer: Any,
        *,
        image_encoder: Callable[[Image.Image], Any] | None = None,
    ) -> dict[str, Any]:
        """Render the current frame and push it into a viewport-like object."""
        payload = self.frame_payload(image_encoder=image_encoder)
        target = getattr(viewer, "widget", viewer)
        target.set_frame_payload(payload)
        return payload

    def _refresh_property_cache(self) -> list[dict[str, Any]]:
        self._refresh_scene_metadata()
        props_by_group = self._scene.list_properties()
        groups: list[dict[str, Any]] = []
        group_index: dict[str, int] = {}
        properties_by_group: dict[str, list[Property]] = {}

        for raw_group_id, props in props_by_group.items():
            group_id = str(raw_group_id)
            prop_list = list(props)
            group_index[group_id] = len(groups)
            properties_by_group[group_id] = prop_list
            for prop in prop_list:
                if prop.param_key is not None and prop.param_key not in self._shared_state:
                    self._register_property_shared_value(prop)
            groups.append(
                {
                    "id": group_id,
                    "title": group_id,
                    "properties": [
                        self._decorate_property_snapshot(group_id, prop, index)
                        for index, prop in enumerate(prop_list)
                    ],
                }
            )

        self._property_groups = groups
        self._property_group_index = group_index
        self._properties_by_group = properties_by_group
        return self._property_groups

    def _capture_current_scene_default_baseline(
        self,
        scene_name: str | None = None,
    ) -> dict[tuple[str, str, str], Any]:
        resolved_scene = self._selected_scene if scene_name is None else scene_name
        if resolved_scene is None:
            return {}
        if resolved_scene in self._default_property_values_by_scene:
            return self._default_property_values_by_scene[resolved_scene]
        baseline: dict[tuple[str, str, str], Any] = {}
        for raw_group_id, props in self._scene.list_properties().items():
            group_id = str(raw_group_id)
            for prop_index, prop in enumerate(list(props)):
                snapshot = _decorate_property_snapshot(
                    _property_snapshot(prop, prop_index)
                )
                baseline[self._property_baseline_key(group_id, snapshot)] = (
                    _copy_shared_value(snapshot.get("value"))
                )
        self._default_property_values_by_scene[resolved_scene] = baseline
        return baseline

    def _default_property_baseline(
        self,
        scene_name: str | None = None,
    ) -> dict[tuple[str, str, str], Any]:
        resolved_scene = self._selected_scene if scene_name is None else scene_name
        if resolved_scene is None:
            return {}
        if (
            resolved_scene not in self._default_property_values_by_scene
            and resolved_scene == self._selected_scene
        ):
            self._capture_current_scene_default_baseline(resolved_scene)
        return self._default_property_values_by_scene.get(resolved_scene, {})

    def _decorate_property_snapshot(
        self,
        group_id: str,
        prop: Property,
        index: int,
    ) -> dict[str, Any]:
        snapshot = _decorate_property_snapshot(_property_snapshot(prop, index))
        baseline = self._default_property_baseline()
        key = self._property_baseline_key(group_id, snapshot)
        default_value = _copy_shared_value(baseline.get(key, snapshot.get("value")))
        snapshot["default_value"] = default_value
        snapshot["is_default"] = bool(
            _shared_values_equal(snapshot.get("value"), default_value)
        )
        snapshot["can_reset"] = bool(key in baseline)
        return snapshot

    def export_residency_snapshot(self) -> ViewerSessionResidencySnapshot:
        """Export lightweight state that survives scene eviction or session rebuild."""
        return ViewerSessionResidencySnapshot(
            shared_state=copy.deepcopy(self._shared_state),
            selected_scene=(
                None if self._selected_scene is None else str(self._selected_scene)
            ),
            default_property_values_by_scene=copy.deepcopy(
                self._default_property_values_by_scene
            ),
        )

    def restore_residency_snapshot(
        self,
        snapshot: ViewerSessionResidencySnapshot | dict[str, Any] | None,
    ) -> "ViewerSession":
        """Restore lightweight persistent viewer state onto the current live scene."""
        if snapshot is None:
            return self
        if isinstance(snapshot, dict):
            snapshot = ViewerSessionResidencySnapshot(
                shared_state=copy.deepcopy(snapshot.get("shared_state", {})),
                selected_scene=(
                    None
                    if snapshot.get("selected_scene") is None
                    else str(snapshot.get("selected_scene"))
                ),
                default_property_values_by_scene=copy.deepcopy(
                    snapshot.get("default_property_values_by_scene", {})
                ),
            )

        self._shared_state = copy.deepcopy(snapshot.shared_state)
        self._default_property_values_by_scene = copy.deepcopy(
            snapshot.default_property_values_by_scene
        )
        self._refresh_scene_metadata()
        self._sync_active_scene_from_shared_state()
        self._restore_or_register_current_camera_slot()
        self._set_committed_camera_state()
        self._refresh_property_cache()
        return self

    def export_live_scenes(self) -> dict[str, Any]:
        """Return all currently live scene objects owned by this session."""
        live_scenes = dict(self._scene_cache)
        if self._selected_scene is not None and self._scene is not None:
            live_scenes[str(self._selected_scene)] = self._scene
        return live_scenes

    def live_scene_names(self) -> list[str]:
        """List scene labels currently backed by live scene objects."""
        return list(self.export_live_scenes().keys())

    def seed_live_scenes(self, scenes: dict[str, Any] | None) -> "ViewerSession":
        """Install pre-created live scenes into this session cache."""
        if not scenes:
            return self
        for scene_name, scene in scenes.items():
            if scene is None:
                continue
            self._scene_cache[str(scene_name)] = scene
        self._cache_current_scene()
        return self

    def evict_live_scene(self, scene_name: str) -> bool:
        """Drop one inactive live scene and release its GL resources."""
        resolved = str(scene_name)
        if resolved == self._selected_scene:
            return False
        scene = self._scene_cache.pop(resolved, None)
        if scene is None:
            return False
        release = getattr(scene, "release", None)
        if callable(release):
            release()
        return True

    def release_live_scenes(self) -> None:
        """Release all live scenes currently owned by this session."""
        seen: set[int] = set()
        for scene in self.export_live_scenes().values():
            if scene is None or id(scene) in seen:
                continue
            seen.add(id(scene))
            release = getattr(scene, "release", None)
            if callable(release):
                release()
        self._scene_cache.clear()

    def _patch_property_cache(
        self,
        group_id: str,
        prop_index: int,
        prop: Property,
    ) -> dict[str, Any]:
        if group_id not in self._property_group_index:
            raise KeyError(f"Unknown property group: {group_id!r}")
        group = self._property_groups[self._property_group_index[group_id]]
        props = group["properties"]
        if not (0 <= prop_index < len(props)):
            raise KeyError(
                f"Unknown property index {prop_index} for group {group_id!r}."
            )
        snapshot = self._decorate_property_snapshot(group_id, prop, prop_index)
        props[prop_index] = snapshot
        return snapshot

    def camera_snapshot(self) -> dict[str, Any]:
        """Collect camera state for frontend adapters."""
        camera_state = self._current_camera_state()
        return {
            "revision": self._revision,
            **camera_state,
            "projection_label": _projection_label(self._camera.perspective_is_on),
            "can_undo": self.can_undo,
            "can_redo": self.can_redo,
            "history_index": 0,
            "history_size": 1,
            "has_pending_delta": self.has_pending_camera_delta(),
        }

    def properties_snapshot(self) -> dict[str, Any]:
        """Collect current scene/property metadata for frontend adapters."""
        return {
            "revision": self._revision,
            "property_groups": copy.deepcopy(self._property_groups),
            "selected_scene": self._selected_scene,
            "available_scenes": list(self._available_scenes),
        }

    def snapshot(self) -> dict[str, Any]:
        """Collect a combined session snapshot for tests and diagnostics."""
        camera_snapshot = self.camera_snapshot()
        properties_snapshot = self.properties_snapshot()
        return {
            "revision": self._revision,
            "camera": camera_snapshot,
            "property_groups": properties_snapshot["property_groups"],
            "selected_scene": properties_snapshot["selected_scene"],
            "available_scenes": properties_snapshot["available_scenes"],
            "summary_text": self.render_snippet(),
        }

    def render_snippet(
        self,
        field_name: str = "field",
        system_name: str = "system",
        topovec_name: str = "tv",
    ) -> str:
        """Generate runnable Python code reproducing the visible render."""
        scene_name = self._selected_scene
        if scene_name is None:
            return (
                "# Current scene is not a built-in topovec.mgl.SCENES scene.\n"
                "# Reproduction snippet is only available for built-in scenes."
            )

        lines = [
            f"ic = {topovec_name}.mgl.render_prepare("
            f"{_format_python_value(scene_name)}, {system_name})",
            f"ic.upload({field_name})",
        ]

        baseline = self._default_property_baseline(scene_name)
        for group in self._property_groups:
            group_id = str(group["id"])
            for prop_snapshot in group["properties"]:
                key = self._property_baseline_key(group_id, prop_snapshot)
                current_value = _copy_shared_value(prop_snapshot.get("value"))
                default_value = baseline.get(key, current_value)
                if _shared_values_equal(current_value, default_value):
                    continue
                lines.append(
                    f"ic.scene[{_format_python_value(group_id)}]"
                    f"[{_format_python_value(prop_snapshot['title'])}] = "
                    f"{_format_python_value(current_value)}"
                )

        lines.append(
            "ic.scene.set_camera("
            f"{format_camera_summary(self._camera, topovec_name=topovec_name)})"
        )
        lines.append("ic.save()")
        return "\n".join(lines)

    def _resolve_property(
        self,
        group_id: str,
        prop_index: int,
        *,
        expected_field: str | None = None,
        expected_title: str | None = None,
    ) -> Property:
        if group_id not in self._properties_by_group:
            raise KeyError(f"Unknown property group: {group_id!r}")
        props = self._properties_by_group[group_id]
        if not (0 <= prop_index < len(props)):
            raise KeyError(
                f"Unknown property index {prop_index} for group {group_id!r}."
            )
        prop = props[prop_index]
        if expected_field is not None and prop.field != expected_field:
            raise RuntimeError(
                f"Property slot mismatch for {group_id!r}[{prop_index}]: "
                f"expected field {expected_field!r}, got {prop.field!r}."
            )
        if expected_title is not None and prop.title != expected_title:
            raise RuntimeError(
                f"Property slot mismatch for {group_id!r}[{prop_index}]: "
                f"expected title {expected_title!r}, got {prop.title!r}."
            )
        return prop

    def apply_camera_delta(
        self,
        *,
        event_type: str,
        dx: float = 0.0,
        dy: float = 0.0,
        wheel: float = 0.0,
    ) -> bool:
        """Apply a live camera delta without mutating explicit-save history."""
        if event_type in {
            "",
            "noop",
            "reset",
            "snap",
            "toggle_perspective",
            "camera_commit",
        }:
            return False

        previous = self._current_camera_state()
        apply_camera_command(
            self._camera,
            event_type=event_type,
            dx=dx,
            dy=dy,
            wheel=wheel,
            rotate_rate=self.rotate_rate,
            roll_rate=self.roll_rate,
            pan_rate=self.pan_rate,
            zoom_rate=self.zoom_rate,
        )
        current = self._current_camera_state()
        return camera_state_signature(current) != camera_state_signature(previous)

    def apply_camera_event(
        self,
        *,
        event_type: str,
        dx: float = 0.0,
        dy: float = 0.0,
        wheel: float = 0.0,
    ) -> bool:
        """Backward-compatible alias for live camera deltas."""
        return self.apply_camera_delta(
            event_type=event_type,
            dx=dx,
            dy=dy,
            wheel=wheel,
        )

    def commit_camera_interaction(self) -> bool:
        """Publish a committed camera pose without tracking history."""
        changed = self.has_pending_camera_delta()
        if changed:
            self._set_committed_camera_state()
            self._increment_revision()
        return changed

    def reset_camera(self) -> bool:
        """Restore the saved home camera for the active profile."""
        previous = self._current_camera_state()
        slot = self._scene_camera_slot()
        home_state = self._shared_state.get(self._home_camera_key(slot))
        if home_state is None:
            home_state = self._ensure_home_state(slot)
        restore_camera(self._camera, home_state)
        current = self._current_camera_state()
        if camera_state_signature(current) == camera_state_signature(previous):
            return False
        self._set_committed_camera_state(current)
        self._increment_revision()
        return True

    def apply_camera_button(self, action: str) -> bool:
        """Apply explicit camera actions."""
        if action in {"undo", "redo"}:
            return False

        if action == "reset":
            return self.reset_camera()

        if action == "snap":
            previous = self._current_camera_state()
            self._camera.snap()
            current = self._current_camera_state()
            if camera_state_signature(current) == camera_state_signature(previous):
                return False
            self._set_committed_camera_state(current)
            self._increment_revision()
            return True

        if action != "toggle_perspective":
            raise ValueError(f"Unsupported camera button action: {action!r}")

        previous = self._current_camera_state()
        self._camera.toggle_perspective()
        current = self._current_camera_state()
        if camera_state_signature(current) == camera_state_signature(previous):
            return False
        self._set_committed_camera_state(current)
        self._increment_revision()
        return True

    def switch_scene(self, scene_name: str) -> bool:
        registry = self._builtin_scene_registry()
        if scene_name not in registry:
            raise KeyError(f"Unknown built-in scene: {scene_name!r}")

        scene_cls = registry[scene_name]
        system = self._scene_system()
        if system is None:
            raise RuntimeError("Scene switching requires `scene.system`.")
        if not _scene_class_supports_system(scene_cls, system):
            raise ValueError(
                f"Scene {scene_name!r} is incompatible with the current system."
            )
        if scene_name == self._selected_scene:
            return False

        previous_slot = self._scene_camera_slot()
        self._store_camera_state(previous_slot)
        self._cache_current_scene()

        scene = self._scene_cache.get(scene_name)
        if scene is None:
            scene = self._create_scene(scene_cls)
            self._scene_cache[scene_name] = scene

        self._attach_scene(scene)
        if self._state is not None:
            self._host.upload(self._state)
        if scene_name not in self._default_property_values_by_scene:
            self._capture_current_scene_default_baseline(scene_name)

        self._sync_active_scene_from_shared_state()
        self._restore_or_register_current_camera_slot()
        self._set_committed_camera_state()
        self._refresh_property_cache()
        self._increment_revision()
        return True

    def set_render_state(self, state: np.ndarray | None) -> "ViewerSession":
        if state is not None:
            self._host.upload(state)
        self._state = state
        self._refresh_property_cache()
        self._increment_revision()
        return self

    def apply_property_input(
        self,
        group_id: str,
        prop_index: int,
        raw_value: Any,
        *,
        expected_field: str | None = None,
        expected_title: str | None = None,
    ) -> UpdateMask:
        prop = self._resolve_property(
            group_id,
            prop_index,
            expected_field=expected_field,
            expected_title=expected_title,
        )
        kind = getattr(prop, "kind", None)
        value = raw_value
        if kind == "numeric":
            snapshot = _decorate_property_snapshot(
                _property_snapshot(prop, prop_index)
            )
            if raw_value is None:
                raise ValueError("Numeric property update requires a value.")
            value = _clamp_numeric_value(float(raw_value), snapshot)
        elif kind == "boolean":
            value = bool(raw_value)
        return self.apply_property_change(
            group_id,
            prop_index,
            value,
            expected_field=expected_field,
            expected_title=expected_title,
        )

    def apply_property_change(
        self,
        group_id: str,
        prop_index: int,
        value: Any,
        *,
        expected_field: str | None = None,
        expected_title: str | None = None,
    ) -> UpdateMask:
        """Apply a scene property update and reconcile any camera side effects."""
        previous_slot = self._scene_camera_slot()
        previous_camera = self._current_camera_state()
        prop = self._resolve_property(
            group_id,
            prop_index,
            expected_field=expected_field,
            expected_title=expected_title,
        )
        with self._host.scene_mutation_scope("apply_property_change"):
            update_mask = UpdateMask(prop.let(value))
        if prop.param_key is not None:
            self._register_property_shared_value(prop)

        if update_mask & UpdateMask.CAMERA_SLOT_CHANGED:
            self._store_camera_state(previous_slot, previous_camera)
            self._restore_or_register_current_camera_slot()

        refresh_mask = UpdateMask(update_mask)
        if update_mask & (
            UpdateMask.REBUILD_PROPERTIES | UpdateMask.REBUILD_PROPERTY_SCHEMA
        ):
            refresh_mask |= self._sync_active_scene_from_shared_state()
            self._refresh_property_cache()
        else:
            self._patch_property_cache(group_id, prop_index, prop)

        self._set_committed_camera_state()
        self._increment_revision()
        return UpdateMask(refresh_mask)


__all__ = [
    "DEFAULT_NUMERIC_STEP",
    "PROJECTION_LABELS",
    "ImageContainerHost",
    "ViewerFrame",
    "ViewerSessionResidencySnapshot",
    "ViewerHost",
    "ViewerSession",
    "_coerce_director_field_state",
    "_decorate_property_snapshot",
    "_effective_numeric_step",
    "_property_snapshot",
    "_quantize_numeric_value",
    "adapt_viewer_host",
    "apply_camera_command",
    "camera_state_signature",
    "format_camera_summary",
    "infer_cubic_system_from_state",
    "restore_camera",
    "serialize_camera",
]



"""Marimo adapter for the reusable TopoVec rendering model.

The marimo layer does not define scenes or renders itself. It binds existing
`topovec.mgl` scenes to anywidget state through :class:`RenderSession` and the
public `inspect(...)` helpers.
"""

from .camera_viewer import (
    DEFAULT_CONTROLS_TEXT,
    CameraViewportWidget,
    DirectorFieldView,
    RenderSession,
    apply_camera_command,
    camera_viewer,
    format_camera_summary,
    inspect,
    inspect_view,
    pil_image_to_data_url,
    restore_camera,
    serialize_camera,
)

__all__ = [
    "DEFAULT_CONTROLS_TEXT",
    "CameraViewportWidget",
    "DirectorFieldView",
    "RenderSession",
    "apply_camera_command",
    "camera_viewer",
    "format_camera_summary",
    "inspect",
    "inspect_view",
    "pil_image_to_data_url",
    "restore_camera",
    "serialize_camera",
]



"""Color-scheme helpers for the `topovec view` Qt desktop app."""

from __future__ import annotations

from dataclasses import dataclass

from PySide6 import QtCore, QtGui, QtWidgets


THEME_MODE_SYSTEM = "system"
THEME_MODE_DARK = "dark"
THEME_MODE_LIGHT = "light"
THEME_MODES = (THEME_MODE_SYSTEM, THEME_MODE_DARK, THEME_MODE_LIGHT)
THEME_SETTING_KEY = "ui_theme"


@dataclass(frozen=True)
class ThemeSpec:
    mode: str
    color_scheme: QtCore.Qt.ColorScheme
    viewport_clear_rgba: tuple[float, float, float, float]
    reset_enabled: QtGui.QColor
    reset_active: QtGui.QColor
    reset_disabled: QtGui.QColor
    reset_background_enabled: str
    reset_background_hover: str
    reset_border_enabled: str
    reset_border_hover: str


def _palette_dark() -> QtGui.QPalette:
    palette = QtGui.QPalette()
    colors = {
        QtGui.QPalette.ColorRole.Window: QtGui.QColor("#1f1f1f"),
        QtGui.QPalette.ColorRole.WindowText: QtGui.QColor("#cccccc"),
        QtGui.QPalette.ColorRole.Base: QtGui.QColor("#181818"),
        QtGui.QPalette.ColorRole.AlternateBase: QtGui.QColor("#252526"),
        QtGui.QPalette.ColorRole.ToolTipBase: QtGui.QColor("#252526"),
        QtGui.QPalette.ColorRole.ToolTipText: QtGui.QColor("#cccccc"),
        QtGui.QPalette.ColorRole.Text: QtGui.QColor("#d4d4d4"),
        QtGui.QPalette.ColorRole.Button: QtGui.QColor("#2d2d30"),
        QtGui.QPalette.ColorRole.ButtonText: QtGui.QColor("#d4d4d4"),
        QtGui.QPalette.ColorRole.BrightText: QtGui.QColor("#ffffff"),
        QtGui.QPalette.ColorRole.Highlight: QtGui.QColor("#264f78"),
        QtGui.QPalette.ColorRole.HighlightedText: QtGui.QColor("#ffffff"),
        QtGui.QPalette.ColorRole.PlaceholderText: QtGui.QColor("#7f848e"),
        QtGui.QPalette.ColorRole.Link: QtGui.QColor("#3794ff"),
        QtGui.QPalette.ColorRole.LinkVisited: QtGui.QColor("#4fc1ff"),
    }
    for role, color in colors.items():
        palette.setColor(role, color)
    palette.setColor(
        QtGui.QPalette.ColorGroup.Disabled,
        QtGui.QPalette.ColorRole.Text,
        QtGui.QColor("#7f848e"),
    )
    palette.setColor(
        QtGui.QPalette.ColorGroup.Disabled,
        QtGui.QPalette.ColorRole.ButtonText,
        QtGui.QColor("#7f848e"),
    )
    palette.setColor(
        QtGui.QPalette.ColorGroup.Disabled,
        QtGui.QPalette.ColorRole.WindowText,
        QtGui.QColor("#9da1a8"),
    )
    return palette


def _palette_light() -> QtGui.QPalette:
    palette = QtGui.QPalette()
    colors = {
        QtGui.QPalette.ColorRole.Window: QtGui.QColor("#f5f7fb"),
        QtGui.QPalette.ColorRole.WindowText: QtGui.QColor("#252526"),
        QtGui.QPalette.ColorRole.Base: QtGui.QColor("#ffffff"),
        QtGui.QPalette.ColorRole.AlternateBase: QtGui.QColor("#eef1f6"),
        QtGui.QPalette.ColorRole.ToolTipBase: QtGui.QColor("#ffffff"),
        QtGui.QPalette.ColorRole.ToolTipText: QtGui.QColor("#252526"),
        QtGui.QPalette.ColorRole.Text: QtGui.QColor("#252526"),
        QtGui.QPalette.ColorRole.Button: QtGui.QColor("#edf1f7"),
        QtGui.QPalette.ColorRole.ButtonText: QtGui.QColor("#252526"),
        QtGui.QPalette.ColorRole.BrightText: QtGui.QColor("#ffffff"),
        QtGui.QPalette.ColorRole.Highlight: QtGui.QColor("#cce8ff"),
        QtGui.QPalette.ColorRole.HighlightedText: QtGui.QColor("#0f1720"),
        QtGui.QPalette.ColorRole.PlaceholderText: QtGui.QColor("#7b8591"),
        QtGui.QPalette.ColorRole.Link: QtGui.QColor("#006ab1"),
        QtGui.QPalette.ColorRole.LinkVisited: QtGui.QColor("#0e639c"),
    }
    for role, color in colors.items():
        palette.setColor(role, color)
    palette.setColor(
        QtGui.QPalette.ColorGroup.Disabled,
        QtGui.QPalette.ColorRole.Text,
        QtGui.QColor("#8a929e"),
    )
    palette.setColor(
        QtGui.QPalette.ColorGroup.Disabled,
        QtGui.QPalette.ColorRole.ButtonText,
        QtGui.QColor("#8a929e"),
    )
    palette.setColor(
        QtGui.QPalette.ColorGroup.Disabled,
        QtGui.QPalette.ColorRole.WindowText,
        QtGui.QColor("#8a929e"),
    )
    return palette


_DARK_SPEC = ThemeSpec(
    mode=THEME_MODE_DARK,
    color_scheme=QtCore.Qt.ColorScheme.Dark,
    viewport_clear_rgba=(0.028, 0.03, 0.035, 1.0),
    reset_enabled=QtGui.QColor("#d8b26e"),
    reset_active=QtGui.QColor("#f0c880"),
    reset_disabled=QtGui.QColor("#596270"),
    reset_background_enabled="rgba(216, 178, 110, 0.18)",
    reset_background_hover="rgba(216, 178, 110, 0.28)",
    reset_border_enabled="rgba(216, 178, 110, 0.32)",
    reset_border_hover="rgba(216, 178, 110, 0.48)",
)

_LIGHT_SPEC = ThemeSpec(
    mode=THEME_MODE_LIGHT,
    color_scheme=QtCore.Qt.ColorScheme.Light,
    viewport_clear_rgba=(0.961, 0.969, 0.984, 1.0),
    reset_enabled=QtGui.QColor("#b0751c"),
    reset_active=QtGui.QColor("#c8892e"),
    reset_disabled=QtGui.QColor("#97a3b0"),
    reset_background_enabled="rgba(200, 137, 46, 0.12)",
    reset_background_hover="rgba(200, 137, 46, 0.2)",
    reset_border_enabled="rgba(176, 117, 28, 0.24)",
    reset_border_hover="rgba(176, 117, 28, 0.36)",
)


def normalize_theme_mode(value: object) -> str:
    text = str(value or THEME_MODE_SYSTEM).strip().lower()
    if text not in THEME_MODES:
        return THEME_MODE_SYSTEM
    return text


def system_color_scheme(app: QtWidgets.QApplication) -> QtCore.Qt.ColorScheme:
    style_hints = app.styleHints()
    scheme = style_hints.colorScheme()
    if scheme == QtCore.Qt.ColorScheme.Unknown:
        palette = app.style().standardPalette()
        return (
            QtCore.Qt.ColorScheme.Dark
            if palette.color(QtGui.QPalette.ColorRole.Window).lightnessF() < 0.5
            else QtCore.Qt.ColorScheme.Light
        )
    return scheme


def theme_spec_for_mode(
    mode: str,
    app: QtWidgets.QApplication,
) -> ThemeSpec:
    normalized = normalize_theme_mode(mode)
    if normalized == THEME_MODE_DARK:
        return _DARK_SPEC
    if normalized == THEME_MODE_LIGHT:
        return _LIGHT_SPEC
    return _DARK_SPEC if system_color_scheme(app) == QtCore.Qt.ColorScheme.Dark else _LIGHT_SPEC


def palette_for_mode(
    mode: str,
    app: QtWidgets.QApplication,
) -> QtGui.QPalette | None:
    normalized = normalize_theme_mode(mode)
    if normalized == THEME_MODE_SYSTEM:
        return None
    return _palette_dark() if normalized == THEME_MODE_DARK else _palette_light()


def apply_theme_mode(
    app: QtWidgets.QApplication,
    mode: str,
) -> ThemeSpec:
    normalized = normalize_theme_mode(mode)
    style_hints = app.styleHints()
    target_scheme = (
        QtCore.Qt.ColorScheme.Unknown
        if normalized == THEME_MODE_SYSTEM
        else theme_spec_for_mode(normalized, app).color_scheme
    )
    if hasattr(style_hints, "setColorScheme"):
        style_hints.setColorScheme(target_scheme)
    palette = palette_for_mode(normalized, app)
    if palette is None:
        app.setPalette(app.style().standardPalette())
    else:
        app.setPalette(palette)
    return theme_spec_for_mode(normalized, app)


__all__ = [
    "THEME_MODE_DARK",
    "THEME_MODE_LIGHT",
    "THEME_MODE_SYSTEM",
    "THEME_MODES",
    "THEME_SETTING_KEY",
    "ThemeSpec",
    "apply_theme_mode",
    "normalize_theme_mode",
    "theme_spec_for_mode",
]



import moderngl as mgl
import numpy as np

from .render import Render
from .buffer import Float3DBuffer
from .solid import ArrowMesh
from .target import RenderTarget
from ..core import NumericProperty, BooleanProperty, ListProperty, PropertiesCollection


#######################################################################################################


class AxesRender(Render):
    VERTEX = """
                #version 330

                const float M_PI = 3.1415926535897932384626433832795;

                uniform mat4 u_view;
                uniform mat4 u_projection;

                uniform float u_ratio;
                uniform float u_length;
                uniform vec2 u_location;

                in vec3 in_origin; // origin of axis
                in vec3 in_dir; // direction of axis

                in vec3 in_pos; // mesh point in mesh coordinates
                in vec3 in_normal; // normal to mesh at in_pos in mesh coordinates
                in vec3 in_color; // color

                out vec3 v_pos; // vertex position in world coordinates
                out vec3 v_normal; // normal to surface in world coordinates
                out vec4 v_color;
                out vec3 v_camera;
                out vec3 v_light_direction;

                mat3 rotmat(vec3 z) {
                    vec3 a = cross(z, vec3(1.0,0.0,0.0));
                    vec3 b = cross(z, vec3(0.0,1.0,0.0));
                    vec3 c = (length(a)>length(b)) ? a : b;
                    vec3 x = normalize(c);
                    vec3 y = cross(x, z);
                    return mat3(x, y, z);
                }

                void main() {
                    vec3 dir = normalize(in_dir);
                    mat3 mesh = rotmat(dir);
                    vec3 pos = in_pos*vec3(u_ratio,u_ratio,1);
                    v_pos = in_origin + mesh * pos;

                    mat4 view = mat4(u_view);
                    mat4 projection = mat4(u_projection);
                    view[0] = normalize(view[0]);
                    view[1] = normalize(view[1]);
                    view[2] = normalize(view[2]);
                    view[3] = vec4(0,0,0,1);
                    projection[2][2] = 1;

                    vec4 glpos = projection * view * vec4(v_pos*u_length, 1.0);
                    gl_Position = glpos/glpos.w + vec4(u_location,0,0);
                    
                    v_normal = mesh * in_normal;

                    v_color = vec4( in_color, 1.0 );

                    v_camera = transpose(view)[2].xyz;
                    v_light_direction = -v_camera;
                } 
            """
    GEOMETRY = None
    FRAGMENT = """
                #version 330

                in vec3 v_normal;
                in vec3 v_pos;
                in vec4 v_color;
                in vec3 v_camera;
                in vec3 v_light_direction;

                out vec4 f_color;

                void main() {
                    vec3 normal = normalize(v_normal);
                    float inner_reflected = dot(reflect(normalize(v_pos-v_camera), normal), v_light_direction);
                    float inner_disperse = dot(normal, v_light_direction);
                    float intensity = 0.4+0.6*clamp(inner_disperse, 0.0, 1.0) + 0.3*clamp(inner_reflected, 0.0, 1.0);
                    f_color = vec4(intensity*v_color.xyz, v_color.t);
                }
            """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.ratio = 0.2
        self.length = 0.1
        self.location_x, self.location_y = -0.8, -0.8
        self._mesh_count = 10

        origin = np.array([[0.55, 0, 0], [0, 0.55, 0], [0, 0, 0.55]], dtype=np.float32)
        self._originBuffer = Float3DBuffer(mglctx=self._ctx, count=origin.shape[0])
        self._originBuffer.upload(origin)

        direction = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float32)
        self._directionBuffer = Float3DBuffer(
            mglctx=self._ctx, count=direction.shape[0]
        )
        self._directionBuffer.upload(direction)

        color = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float32)
        self._colorBuffer = Float3DBuffer(mglctx=self._ctx, count=color.shape[0])
        self._colorBuffer.upload(color)

        self.init_vao()

    def _build_mesh(self):
        return ArrowMesh(mglctx=self._ctx, count=self._mesh_count)

    def _build_vao(self, mesh):
        return self._ctx.vertex_array(
            self._prog,
            [
                (mesh.pos.vbo, "3f", "in_pos"),
                (mesh.normal.vbo, "3f", "in_normal"),
                (self._directionBuffer.vbo, "3f/i", "in_dir"),
                (self._originBuffer.vbo, "3f/i", "in_origin"),
                (self._colorBuffer.vbo, "3f/i", "in_color"),
            ],
            mesh.idx.ibo,
        )

    def init_vao(self):
        self._replace_owned_geometry(self._build_mesh, self._build_vao)

    def render(self, scales=None, modelview=None, projection=None, target: RenderTarget | None = None):
        render_target = (
            RenderTarget.from_current(self._ctx)
            if target is None
            else target
        )
        framebuffer = render_target.fbo
        previous_color_mask = framebuffer.color_mask
        previous_depth_mask = framebuffer.depth_mask
        try:
            render_target.clear_depth()
            framebuffer.color_mask = True, True, True, True
            framebuffer.use()  # Help with buggy driver.
            self._ctx.enable_only(mgl.DEPTH_TEST)

            self.u_view.value = tuple(modelview.T.flatten())
            self.u_projection.value = tuple(projection.T.flatten())

            self.u_ratio.value = self.ratio
            self.u_length.value = self.length
            self.u_location.value = (self.location_x, self.location_y)

            self._vao.render(mode=mgl.TRIANGLES, instances=3, vertices=-1, first=0)
        finally:
            framebuffer.color_mask = previous_color_mask
            framebuffer.depth_mask = previous_depth_mask

    def list_properties(self):
        return [
            NumericProperty(self, "location_x", title="X", min=-1.0, max=1.0, count=200),
            NumericProperty(self, "location_y", title="Y", min=-1.0, max=1.0, count=200),
            NumericProperty(self, "length", title="Length", min=0.0, max=1.0, count=200),
            NumericProperty(self, "ratio", title="Ratio", min=0.01, max=1.0, count=100),
        ]
